from os import path
from ast import literal_eval
from collections import defaultdict

from luna.mol.entry import REGEX_RESNUM_ICODE
from luna.util.config import Config
from luna.util.exceptions import IllegalArgumentError


AROMATIC_STACKINGS = ["pi-stacking",
                      "face-to-face pi-stacking",
                      "face-to-edge pi-stacking",
                      "face-to-slope pi-stacking",
                      "edge-to-edge pi-stacking",
                      "edge-to-face pi-stacking",
                      "edge-to-slope pi-stacking",
                      "displaced face-to-face pi-stacking",
                      "displaced face-to-edge pi-stacking",
                      "displaced face-to-slope pi-stacking"]


class InteractionFilter:
    """ Filter interactions based on their components.

    Parameters
    ----------
    ignore_self_inter : bool
        If True, ignore interactions involving atoms of the same compound.
    ignore_intra_chain : bool
        If True, ignore intra-chain interactions (e.g., interactions
        between residues in the same protein chain).
    ignore_inter_chain : bool
        If True, ignore interactions between different chains.
    ignore_res_res : bool
        If True, ignore residue-residue interactions.
    ignore_res_nucl : bool
        If True, ignore residue-nucleotide interactions.
    ignore_res_hetatm : bool
        If True, ignore residue-ligand interactions.
    ignore_nucl_nucl : bool
        If True, ignore nucleotide-nucleotide interactions.
    ignore_nucl_hetatm : bool
        If True, ignore nucleotide-ligand interactions.
    ignore_hetatm_hetatm : bool
        If True, ignore ligand-ligand interactions.
    ignore_h2o_h2o : bool
        If True, ignore water-water interactions.
    ignore_any_h2o : bool
        If True, ignore all interactions involving water.
    ignore_multi_comps : bool
        If True, ignore interactions established by atom groups composed of
        multiple compounds (e.g.: amides formed by peptide bonds involve
        two residues).
    ignore_mixed_class : bool
        If True, ignore interactions established by atom groups comprising
        mixed compound classes (e.g. residues and ligands bound by a
        covalent bond).
    """

    def __init__(self, ignore_self_inter=True, ignore_intra_chain=True,
                 ignore_inter_chain=True, ignore_res_res=True,
                 ignore_res_nucl=True, ignore_res_hetatm=True,
                 ignore_nucl_nucl=True, ignore_nucl_hetatm=True,
                 ignore_hetatm_hetatm=True, ignore_h2o_h2o=True,
                 ignore_any_h2o=False, ignore_multi_comps=False,
                 ignore_mixed_class=False):

        self.ignore_self_inter = ignore_self_inter
        self.ignore_intra_chain = ignore_intra_chain
        self.ignore_inter_chain = ignore_inter_chain
        self.ignore_res_res = ignore_res_res
        self.ignore_res_nucl = ignore_res_nucl
        self.ignore_res_hetatm = ignore_res_hetatm
        self.ignore_nucl_nucl = ignore_nucl_nucl
        self.ignore_nucl_hetatm = ignore_nucl_hetatm
        self.ignore_hetatm_hetatm = ignore_hetatm_hetatm
        self.ignore_h2o_h2o = ignore_h2o_h2o
        self.ignore_any_h2o = ignore_any_h2o
        self.ignore_multi_comps = ignore_multi_comps
        self.ignore_mixed_class = ignore_mixed_class

    @classmethod
    def new_pli_filter(cls, ignore_res_hetatm=False,
                       ignore_hetatm_hetatm=False, ignore_any_h2o=False,
                       ignore_self_inter=False, **kwargs):
        """Initialize the default filter for protein-ligand interactions.

        Returns
        -------
         : `InteractionFilter`
        """
        return cls(ignore_res_hetatm=ignore_res_hetatm,
                   ignore_hetatm_hetatm=ignore_hetatm_hetatm,
                   ignore_any_h2o=ignore_any_h2o,
                   ignore_self_inter=ignore_self_inter, **kwargs)

    @classmethod
    def new_ppi_filter(cls, ignore_res_res=False, ignore_inter_chain=False,
                       ignore_intra_chain=False, ignore_any_h2o=False,
                       ignore_self_inter=False, **kwargs):
        """Initialize the default filter for protein-protein interactions.

        Returns
        -------
         : `InteractionFilter`
        """
        return cls(ignore_res_res=ignore_res_res,
                   ignore_inter_chain=ignore_inter_chain,
                   ignore_intra_chain=ignore_intra_chain,
                   ignore_any_h2o=ignore_any_h2o,
                   ignore_self_inter=ignore_self_inter, **kwargs)

    @classmethod
    def new_pni_filter(cls, ignore_res_nucl=False, ignore_inter_chain=False,
                       ignore_intra_chain=False, ignore_any_h2o=False,
                       ignore_self_inter=False, **kwargs):
        """Initialize the default filter for protein-nucleotide interactions.

        Returns
        -------
         : `InteractionFilter`
        """
        return cls(ignore_res_nucl=ignore_res_nucl,
                   ignore_inter_chain=ignore_inter_chain,
                   ignore_intra_chain=ignore_intra_chain,
                   ignore_any_h2o=ignore_any_h2o,
                   ignore_self_inter=ignore_self_inter, **kwargs)

    @classmethod
    def new_nni_filter(cls, ignore_nucl_nucl=False, ignore_inter_chain=False,
                       ignore_intra_chain=False, ignore_any_h2o=False,
                       ignore_self_inter=False, **kwargs):
        """Initialize the default filter for nucleotide-nucleotide
        interactions.

        Returns
        -------
         : `InteractionFilter`
        """
        return cls(ignore_nucl_nucl=ignore_nucl_nucl,
                   ignore_inter_chain=ignore_inter_chain,
                   ignore_intra_chain=ignore_intra_chain,
                   ignore_any_h2o=ignore_any_h2o,
                   ignore_self_inter=ignore_self_inter, **kwargs)

    @classmethod
    def new_nli_filter(cls, ignore_nucl_hetatm=False,
                       ignore_hetatm_hetatm=False, ignore_any_h2o=False,
                       ignore_self_inter=False, **kwargs):
        """Initialize the default filter for nucleotide-ligand interactions.

        Returns
        -------
         : `InteractionFilter`
        """
        return cls(ignore_nucl_hetatm=ignore_nucl_hetatm,
                   ignore_hetatm_hetatm=ignore_hetatm_hetatm,
                   ignore_any_h2o=ignore_any_h2o,
                   ignore_self_inter=ignore_self_inter, **kwargs)

    @classmethod
    def from_config_file(cls, config_file):
        """Initialize from a configuration file.

        Parameters
        ----------
        config_file : str
            The configuration file pathname.

        Returns
        -------
         : `InteractionFilter`
        """

        if not path.exists(config_file):
            raise OSError("File '%s' does not exist." % config_file)

        params = Config(config_file)

        inter_filter = None
        if "default" in params.sections():
            params_dict = params.get_section_map("default")
            filter_type = params_dict.pop("type", None)

            if filter_type is not None:
                filter_type = filter_type.upper()

                available_filters = ["PLI", "PPI", "PNI", "NNI", "NLI"]
                if filter_type not in available_filters:
                    raise KeyError("The accepted default filter types are: "
                                   "%s." % ", ".join(available_filters))

                if filter_type == "PLI":
                    inter_filter = cls.new_pli_filter()

                elif filter_type == "PPI":
                    inter_filter = cls.new_ppi_filter()

                elif filter_type == "PNI":
                    inter_filter = cls.new_pni_filter()

                elif filter_type == "NNI":
                    inter_filter = cls.new_nni_filter()

                elif filter_type == "NLI":
                    inter_filter = cls.new_nli_filter()

        if "ignore" in params.sections():
            params_dict = params.get_section_map("ignore")
            params_dict = {"ignore_" + k: v
                           for k, v in params_dict.items()}

            if inter_filter:
                for prop, val in params_dict.items():
                    setattr(inter_filter, prop, val)
            else:
                inter_filter = cls(**params_dict)

        return inter_filter

    def is_valid_pair(self, src_grp, trgt_grp):
        """Evaluate if a pair of atom groups are valid according to the flags
        defined in this class.

        src_grp, trgt_grp : :class:`luna.mol.groups.AtomGroup`
        """

        # It will always ignore interactions involving the same atom groups.
        # Loops in the graph is not permitted and does not make any sense.
        if src_grp == trgt_grp:
            return False

        # It will always ignore interactions involving atoms and the group
        # to which they belong to. For example, the centroid of an aromatic
        # ring cannot interact with an atom that belongs to the ring.
        # It is a type of Loop.
        if src_grp.contain_group(trgt_grp) or trgt_grp.contain_group(src_grp):
            return False

        # If one of the groups contain atoms from different compounds.
        has_multi_comps = (len(src_grp.compounds) > 1
                           or len(trgt_grp.compounds) > 1)
        if self.ignore_multi_comps and has_multi_comps:
            return False

        # If one of the groups contain compounds from different classes as,
        # for instance, residue and ligand. It means that compounds from
        # different classes are covalently bonded to each other.
        has_any_mixed = (src_grp.is_mixed() or trgt_grp.is_mixed())
        if self.ignore_mixed_class and has_any_mixed:
            return False

        # It ignores interactions involving the same compounds if required.
        # As each group may have atoms from different compounds, we can check
        # if there is at least one common compound between the two groups.
        # Remember that if two or more compounds exist in a group, it means
        # that these compounds are covalently bonded and should be considered
        # the same compound.
        # For example: a carbohydrate can be expressed in a PDB as its
        # subparts:
        #      E.g.: GLC + GLC = CBI
        # The same applies to any group formed after covalently bonding a
        # residue to a hetatm (ligand or non-standard amino acid
        # represented as hetatm)
        is_same_compounds = \
            len(src_grp.compounds.intersection(trgt_grp.compounds)) >= 1
        if self.ignore_self_inter and is_same_compounds:
            return False

        # Check if two groups contain the same chains and if both of them
        # contain only one chain. The second condition removes groups
        # containing residues of different chains as may occur due to
        # disulfide bonds. Note, however, that this flag will be used only
        # as a filter for intra-interactions in protein/RNA/DNA chains.
        is_same_chain = (src_grp.get_chains() == trgt_grp.get_chains()
                         and len(src_grp.get_chains()) == 1)

        # Filters for residue-residue interactions if required.
        is_res_res = (src_grp.is_residue() and trgt_grp.is_residue())
        if is_res_res:
            # Ignore all residue-residue interactions.
            if self.ignore_res_res:
                return False
            # Ignore all intra-chain interactions involving two residues.
            elif self.ignore_intra_chain and is_same_chain:
                return False
            elif self.ignore_inter_chain and not is_same_chain:
                return False

        # It ignores residue-nucleic acid interactions if required.
        is_res_nucl = ((src_grp.is_residue() and trgt_grp.is_nucleotide())
                       or (src_grp.is_nucleotide() and trgt_grp.is_residue()))
        if self.ignore_res_nucl and is_res_nucl:
            return False

        # It ignores residue-ligand interactions if required.
        is_res_hetatm = ((src_grp.is_residue() and trgt_grp.is_hetatm())
                         or (src_grp.is_hetatm() and trgt_grp.is_residue()))
        if self.ignore_res_hetatm and is_res_hetatm:
            return False

        # Filters for nucleic acid-nucleic acid interactions if required.
        is_nucl_nucl = (src_grp.is_nucleotide() and trgt_grp.is_nucleotide())
        if is_nucl_nucl:
            # Ignore all nucleic acid-nucleic acid interactions
            if self.ignore_nucl_nucl:
                return False
            # Ignore all intra-chain interactions involving
            # two nucleic acids (RNA/DNA chains).
            elif self.ignore_intra_chain and is_same_chain:
                return False
            # Ignore all inter-chain interactions involving
            # two nucleic acids (RNA/DNA chains).
            elif self.ignore_inter_chain and not is_same_chain:
                return False

        # It ignores nucleic acid-ligand interactions if required.
        is_nucl_hetatm = ((src_grp.is_nucleotide() and trgt_grp.is_hetatm())
                          or (src_grp.is_hetatm()
                              and trgt_grp.is_nucleotide()))
        if self.ignore_nucl_hetatm and is_nucl_hetatm:
            return False

        # It ignores ligand-ligand interactions if required.
        is_hetatm_hetatm = (src_grp.is_hetatm() and trgt_grp.is_hetatm())
        if self.ignore_hetatm_hetatm and is_hetatm_hetatm:
            return False

        # It ignores interactions of other compound types with water.
        # It enables the possibility of identifying water-bridged interaction.
        # Eg: residue-water, ligand-water = residue -- water -- ligand.
        is_any_h2o = (src_grp.is_water() or trgt_grp.is_water())
        if self.ignore_any_h2o and is_any_h2o:
            return False

        # It ignores interactions involving two waters if required.
        # If ON, it will produce water-bridged interactions of multiple levels
        # E.g.: residue -- h2o -- h2o -- ligand
        #       residue -- residue -- h2o -- h2o -- ligand.
        is_h2o_h2o = (src_grp.is_water() and trgt_grp.is_water())
        if self.ignore_h2o_h2o and is_h2o_h2o:
            return False

        return True

    def save_config_file(self, config_file):
        """Save the interaction filter parameters into a configuration file.

        Parameters
        ----------
        config_file : str
            The output configuration file.
        """
        with open(config_file, "w") as OUT:
            OUT.write("[ignore]\n")
            OUT.write("self_inter = %s\n" % self.ignore_self_inter)
            OUT.write("intra_chain = %s\n" % self.ignore_intra_chain)
            OUT.write("inter_chain = %s\n" % self.ignore_inter_chain)
            OUT.write("res_res = %s\n" % self.ignore_res_res)
            OUT.write("res_nucl = %s\n" % self.ignore_res_nucl)
            OUT.write("res_hetatm = %s\n" % self.ignore_res_hetatm)
            OUT.write("nucl_nucl = %s\n" % self.ignore_nucl_nucl)
            OUT.write("nucl_hetatm = %s\n" % self.ignore_nucl_hetatm)
            OUT.write("hetatm_hetatm = %s\n" % self.ignore_hetatm_hetatm)
            OUT.write("h2o_h2o = %s\n" % self.ignore_h2o_h2o)
            OUT.write("any_h2o = %s\n" % self.ignore_any_h2o)
            OUT.write("multi_comps = %s\n" % self.ignore_multi_comps)
            OUT.write("mixed_class = %s\n" % self.ignore_mixed_class)


class BindingModeCondition:

    """Define binding mode conditions to filter interactions.

    Parameters
    ----------

    condition : str
        A string defining which chains, compounds, or atoms should be accepted.
        If ``condition`` is the wildcard '*', then all chains, compounds, and
        atoms will be considered valid. Otherwise, ``condition`` should have
        the format '<CHAIN ID>/<COMPOUND NAME>/<COMPOUND NUMBER>/<ATOM>'.
        Wildcards are accepted for each one of these fields.
        For example:

            * '\\*/HIS/\\*/\\*': represents all histidines' \
                                 atoms from all chains.
            * 'A/CBL/\\*/\\*' represents all ligands named CBL \
                            from chain A.
            * 'B/HIS/\\*/N\\*' represents all histidines' nitrogens \
                             from chain B.

    Attributes
    ----------
    accept_all : bool
        If True, all chains, compounds, and atoms will be considered valid.
    accept_all_chains : bool
        If True, all chains will be considered valid.
    accept_all_comps : bool
        If True, all compound names will be considered valid.
    accept_all_comp_nums : bool
        If True, all compound numbers (residue sequence number in the PDB
        format) will be considered valid.
    accept_all_atoms : bool
        If True, all atoms will be considered valid.
    chain_id : str or None
        If provided, accept only chains whose id matches ``chain_id``.
    comp_name : str or None
        If provided, accept only compounds whose name matches ``comp_name``.
    comp_num : int or None
        If provided, accept only compounds whose sequence number matches
        ``comp_num``.
    comp_icode : str or None
        If provided, accept only compounds whose insertion code matches
        ``comp_icode``.
    atom : str or None
        If provided, accept only atoms whose name matches ``atom``.
    """

    def __init__(self, condition):
        self.accept_all = False
        self.accept_all_chains = False
        self.accept_all_comps = False
        self.accept_all_comp_nums = False
        self.accept_all_atoms = False

        self.chain_id = None
        self.comp_name = None
        self.comp_num = None
        self.comp_icode = None
        self.atom = None

        self._condition_repr = condition

        if condition is not None:
            self._parse_condition(condition.upper())

    def _parse_condition(self, condition):
        # Accept everything.
        if condition == "*":
            self.accept_all = True
        else:
            chain, comp_name, comp_num, atom = condition.split("/")

            if chain == "*":
                self.accept_all_chains = True
            else:
                self.chain_id = chain

            if comp_name == "*":
                self.accept_all_comps = True
            else:
                self.comp_name = comp_name

            if comp_num == "*":
                self.accept_all_comp_nums = True
            else:
                # Separate ligand number from insertion code.
                matched = REGEX_RESNUM_ICODE.match(comp_num)
                if matched:
                    comp_num = matched.group(1)
                    try:
                        assert float(comp_num).is_integer()
                        comp_num = int(comp_num)
                    except (ValueError, AssertionError):
                        error_msg = ("The informed compound number '%s' is "
                                     "invalid. It must be an integer."
                                     % str(comp_num))
                        raise IllegalArgumentError(error_msg)

                    icode = (None if matched.group(2) == ""
                             else matched.group(2))
                else:
                    error_msg = ("The compound number and its insertion code "
                                 "(if applicable) '%s' is invalid. It must be "
                                 "an integer followed by one insertion code "
                                 "character when applicable." % comp_num)
                    raise IllegalArgumentError(error_msg)

                self.comp_num = comp_num
                self.comp_icode = icode

            if atom == "*":
                self.accept_all_atoms = True
            else:
                self.atom = atom

    def is_valid(self, atm_grp):
        """Check if an atom group is valid or not based on this condition.

        atm_grp : :class:`luna.mol.groups.AtomGroup`
        """

        # Accept everything.
        if self.accept_all:
            return True

        # Accept everything.
        if (self.accept_all_chains and self.accept_all_comps
                and self.accept_all_comp_nums and self.accept_all_atoms):
            return True

        # Tries to identify the first valid compound in the atom group.
        for atm in atm_grp.atoms:
            comp = atm.parent

            is_chain_valid = self.accept_all_chains
            is_comp_valid = self.accept_all_comps
            is_comp_num_valid = self.accept_all_comp_nums
            is_atom_valid = self.accept_all_atoms

            if self.chain_id is not None and self.chain_id == comp.parent.id:
                is_chain_valid = True

            if self.comp_name is not None and self.comp_name == comp.resname:
                is_comp_valid = True

            if self.comp_num is not None and self.comp_num == comp.id[1]:
                icode = None if comp.id[2].strip() == "" else comp.id[2]

                if self.comp_icode == icode:
                    is_comp_num_valid = True

            if self.atom is not None:
                # Verify element equality.
                if self.atom.endswith("*"):
                    elem = self.atom.rstrip("*")

                    if elem == atm.element:
                        is_atom_valid = True

                # Verify atom name equality.
                elif self.atom == atm.name:
                    is_atom_valid = True

            if (is_chain_valid and is_comp_valid
                    and is_comp_num_valid and is_atom_valid):
                return True

        return False

    def __repr__(self):
        return "<BindingModeCondition: %s>" % self._condition_repr


class BindingModeFilter:

    """Filter interactions based on a set of binding mode conditions.

    Parameters
    ----------
    config : dict of {str : iterable}
        A dict defining binding modes and how interactions should be validated.
        Each key represents an interaction type and values are an iterable of
        `BindingModeCondition` instances.
    """

    def __init__(self, config, names_map=None):
        self.config = config
        self.names_map = names_map or {}

    @classmethod
    def from_config_file(cls, config_file):
        """Initialize from a configuration file.

        Parameters
        ----------
        config_file : str
            The configuration file pathname.

        Returns
        -------
         : `BindingModeFilter`

        Examples
        --------

        It follows an example of a configuration file::

            ; To configurate an interaction type, create a new line and define
            ; the interaction: [New interaction]. Then you can define whether
            ; or not all interactions must be accepted by setting 'accept_only'
            ; to True or False.

            ; If you want to specify binding modes, use the variable
            ; 'accept_only', which expects a list of strings in the format:
            ; <CHAIN ID>/<COMPOUND NAME>/<COMPOUND NUMBER>/<ATOM>.
            ;
            ; Wildcards are accepted for the expected fields.
            ; For example, "*/HIS/*/*" represents all histidines' atoms from \
all chains.
            ;               "A/CBL/*/*" represents all ligands named CBL from \
chain A.
            ;               "B/HIS/*/N*" represents all histidines' nitrogens \
from chain B.

            [Hydrogen bond]
            accept_only = ["A/LYS/245/*", "*/HIS/*/*"]

            [Hydrophobic]
            accept_all = True

            [Cation-pi]
            accept_only = ["*"]
            accept_all = True

            [Weak hydrogen bond]
            accept_all = False
            accept_only = ["*/THR/434/O*"]

            [Face-to-edge pi-stacking]
            accept_only = ["*"]

            [Aromatic stacking]
            accept_all = True

            [*]
            accept_all = False
        """

        if not path.exists(config_file):
            raise OSError("File '%s' does not exist." % config_file)

        filtering_config = {}
        names_mapping = {}

        config = Config(config_file)
        for inter_type in config.sections():
            params = config.get_section_map(inter_type)

            ori_name = inter_type
            inter_type = inter_type.lower()

            accept_all = False
            values = []
            if "accept_all" in params:
                accept_all = literal_eval(params["accept_all"])
                if accept_all is True:
                    values = ["*"]
                else:
                    values = [None]

            elif "accept_only" in params and not accept_all:
                values = literal_eval(params["accept_only"])

                if "*" in values:
                    values = ["*"]

            conditions = [BindingModeCondition(condition)
                          for condition in values]

            filtering_config[inter_type] = conditions
            names_mapping[inter_type] = ori_name

        return cls(filtering_config, names_mapping)

    def is_valid(self, inter):
        """Check if an interaction is valid or not based on this binding mode
        configuration.

        inter : :class:`luna.interaction.type.InteractionType`
        """
        if inter.type.lower() in self.config:
            inter_type = inter.type.lower()

        elif ("aromatic stacking" in self.config
                and inter.type.lower() in AROMATIC_STACKINGS):
            inter_type = "aromatic stacking"

        elif "*" in self.config:
            inter_type = "*"

        else:
            return False

        for condition in self.config[inter_type]:
            is_src_grp_valid = condition.is_valid(inter.src_grp)
            is_trgt_grp_valid = condition.is_valid(inter.trgt_grp)

            if is_src_grp_valid or is_trgt_grp_valid:
                return True

        return False

    def save_config_file(self, config_file):
        """Save the binding mode filter parameters into a
        configuration file.

        Parameters
        ----------
        config_file : str
            The output configuration file.
        """
        args_by_section = defaultdict(list)

        for inter, conditions in self.config.items():
            section_name = self.names_map.get(inter, inter)

            multi_conditions = []
            for condition in conditions:
                if (condition.accept_all is False
                        and condition._condition_repr is None):
                    args_by_section[section_name].append(("accept_all", False))

                elif condition.accept_all is True:
                    args_by_section[section_name].append(("accept_all", True))

                else:
                    multi_conditions.append(condition._condition_repr)

            if multi_conditions:
                args_by_section[section_name].append(("accept_only",
                                                      str(multi_conditions)))

        with open(config_file, "w") as OUT:
            for k in args_by_section:
                OUT.write("[%s]\n" % k)
                for arg, val in args_by_section[k]:
                    OUT.write("%s = %s\n" % (arg, val))
                OUT.write("\n")



"""
    Pattern: MAJOR.MINOR.PATCH as in https://semver.org/, where
       - MAJOR stands for big scientific updates;

       - MINOR stands for updates that "break" the compatibility
            between pharmacophoric perception, interactions,
            and interaction fingerprints;

       - PATCH stands for backward compatible bug fixes.
"""
version_info = (0, 14, 0)
version = '.'.join(str(c) for c in version_info)
__version__ = version


def has_version_compatibility(v):
    version_to_check = v
    if isinstance(v, str):
        version_to_check = [int(i) for i in v.split(".")]
    try:
        if tuple(version_to_check[0:2]) == version_info[0:2]:
            return True
        else:
            return False
    except TypeError:
        raise



import logging
from io import StringIO

from Bio.PDB.PDBIO import Select

from luna.pdb.io.base import PDBIO
from luna.util.exceptions import FileNotCreated


logger = logging.getLogger()


def save_to_file(entity, 
                 output_file, 
                 select=Select(), 
                 write_conects=True,
                 write_end=True, 
                 preserve_atom_numbering=True, 
                 sort=False):
    """
    Write a structure, model, chain, residue, or atom object to a PDB file.

    Parameters
    ----------
    entity : :class:`~luna.pdb.core.entity.Entity`
         The structure, model, chain, residue, or atom to be written.
    output_file : str
        Path to the output PDB file.
    select : :class:`Bio.PDB.PDBIO.Select`
        Selector for filtering atoms/residues/chains. 
        Defaults to selecting all.
    write_conects : bool
        If True, write CONECT records (if available).
    write_end : bool
        If True, write the END record.
    preserve_atom_numbering : bool
        If True, keep the original atom serial numbers.
        If False, atoms are renumbered sequentially.
    sort : bool, default=False
        If True, sort chains and residues before writing.
    """
    try:
        io = PDBIO()
        io.set_structure(entity)
        io.save(output_file, 
                select=select,
                write_conects=write_conects,
                write_end=write_end,
                preserve_atom_numbering=preserve_atom_numbering,
                sort=sort)

    except Exception as e:
        logger.exception(e)
        raise FileNotCreated(f"PDB file '{output_file}' could not be created.")


def entity_to_string(entity, 
                     select=Select(),
                     write_conects=True,
                     write_end=True,
                     preserve_atom_numbering=True):
    """
    Convert a structure, model, chain, residue, or atom to string.

    This function works on a structural level. That means if ``entity`` is not
    a :class:`~luna.pdb.core.structure.Structure` object, the structure will
    be recovered directly from ``entity``. Therefore, use ``select`` to select
    specific chains, residues, and atoms from the structure object.

    Parameters
    ----------
    entity : :class:`~luna.pdb.core.entity.Entity`
        The object to convert.
    select : :class:`Bio.PDB.PDBIO.Select`
        Selector for filtering atoms/residues/chains. 
        Defaults to selecting all.
    write_conects : bool
        If True, write CONECT records.
    write_end : bool
        If True, write the END record.
    preserve_atom_numbering : bool
        If True, keep the original atom serial numbers.
        If False, atoms are renumbered sequentially.

    Returns
    -------
    str
        The serialized PDB content as a string.
    """
    fh = StringIO()
    io = PDBIO()
    io.set_structure(entity)
    io.save(fh, 
            select=select,
            write_conects=write_conects,
            write_end=write_end,
            preserve_atom_numbering=preserve_atom_numbering)
    return fh.getvalue()



from os import remove
import math
import glob
import subprocess
import os.path
import warnings

from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from Bio.Align import MultipleSeqAlignment
from Bio.Alphabet import generic_protein

from luna.pdb.parser.helpers import load_from_file
from luna.pdb.io.helpers import save_to_file

from luna.util.exceptions import InvalidSuperpositionFileError
from luna.util.file import is_directory_valid


import logging
logger = logging.getLogger()


TMALIGN = "/bin/tmalign"


class TMAlignment(MultipleSeqAlignment):

    """ Store TM-align results, including sequence alignment and TM-score.

    Parameters
    ----------
    score : float
        TM-score.
    records : iterable of :class:`Bio.SeqRecord.SeqRecord`
        Same length protein sequences. This may be an empty list.
    **kwargs : dict, optional
        Extra arguments to `TMAlignment`.
        Refer to :class:`Bio.Align.MultipleSeqAlignment` documentation for a
        list of all possible arguments.
    """

    def __init__(self, score, records, **kwargs):
        self.score = score
        super().__init__(records, **kwargs)


def align_structures(pdb_to_align, ref_pdb, output_path=None, tmalign=None):
    """Align two PDB files with TM-align.

    .. warning::

        TM-align performs pair-wise structural alignments.
        Therefore, if your PDB file contains multiple structures, it is
        recommended you extract the chains first and align them separately,
        otherwise the alignment may not produce the expected results.

        To extract chains you can use :class:`~luna.pdb.io.extractor.Extractor`.

    Parameters
    ----------
    pdb_to_align : str
        The PDB structure that will be aligned to ``ref_pdb``.
    ref_pdb : str
        Reference PDB file, i.e., align ``pdb_to_align`` to ``ref_pdb``.
    output_path : str
        Where to save TM-align output structures.
    tmalign : str
        Pathname to TM-align binary. The default value is '/bin/tmalign'.

    Returns
    -------
     : `TMAlignment`
     Results of TM-align, including sequence alignment and TM-score.

    Examples
    --------

    >>> from luna.util.default_values import LUNA_PATH
    >>> from luna.align.tmalign import align_structures
    >>> pdb1 = "{LUNA_PATH}/tutorial/inputs/3QQF.pdb"
    >>> pdb2 = "{LUNA_PATH}/tutorial/inputs/3QQK.pdb"
    >>> alignment = align_structures(pdb1, pdb2, "./", \
tmalign="/media/data/Workspace/Softwares/TMalignc/TMalign")
    >>> print(alignment.score)
    0.98582
    """

    tmalign = tmalign or TMALIGN

    logger.info("TM-align will try to align the files %s and %s."
                % (pdb_to_align, ref_pdb))

    output = _run_tmalign(pdb_to_align, ref_pdb, output_path, tmalign)

    seq_pair, tm_score = get_seq_records(output, pdb_to_align, ref_pdb)

    alignment = TMAlignment(tm_score,
                            records=seq_pair,
                            alphabet=generic_protein)

    logger.info("TM-align finished successfully. %s."
                % alignment[0].description)

    return alignment


def _run_tmalign(file1, file2, output_path, tmalign):

    logger.debug("It will try to execute the command: '%s %s %s'."
                 % (tmalign, file1, file2))

    for fname in (file1, file2):
        if not os.path.isfile(fname):
            logging.error("Missing file: %s", fname)
            raise FileNotFoundError("Missing file: %s", fname)

    try:
        if output_path is not None and output_path.strip() != "":
            if is_directory_valid(output_path):
                logger.debug("The superposition files will be saved "
                             "at the directory '%s'" % output_path)

                filename = os.path.split(os.path.basename(file1))[1]
                output_file = "%s/%s.sup" % (output_path, filename)
                args = [tmalign, file1, file2, "-o", output_file]
        else:
            args = [tmalign, file1, file2]

        output = subprocess.check_output(args)
    except subprocess.CalledProcessError as e:
        logger.exception(e)

        raise RuntimeError("TMalign failed for PDB files: %s and %s"
                           % (file1, file2))

    return output.decode()


def get_seq_records(tm_output, aligned_pdb_id, ref_pdb_id):
    """Create a pair of :class:`Bio.SeqRecord.SeqRecord` from a
    TM-align output.

    Parameters
    ----------
    tm_output : str
        The output produced by TM-align.
    aligned_pdb_id : str
        An identifier for the aligned PDB file.
    ref_pdb_id : str
        An identifier for the reference PDB file.
    """
    logger.debug("Parsing the TMalign output.")

    lines = tm_output.splitlines()

    # Extract the TM-score (measure of structure similarity)
    # Take the mean of the (two) given TM-scores -- not sure which is reference
    tm_scores = []
    for line in lines:
        if line.startswith('TM-score'):
            # TMalign v. 2012/05/07 or earlier
            tm_scores.append(float(line.split(None, 2)[1]))
        elif 'TM-score=' in line:
            # TMalign v. 2013/05/11 or so
            tokens = line.split()
            for token in tokens:
                if token.startswith('TM-score='):
                    _key, _val = token.split('=')
                    tm_scores.append(float(_val.rstrip(',')))
                    break

    tm_score = math.fsum(tm_scores) / len(tm_scores)
    # Extract the sequence alignment
    last_lines = lines[-7:]

    assert last_lines[0].startswith('(":"')  # (":" denotes the residues pairs
    assert last_lines[-1].startswith('Total running time is')

    aligned_seq, ref_seq = last_lines[1].strip(), last_lines[3].strip()

    return (SeqRecord(Seq(aligned_seq),
                      id=aligned_pdb_id,
                      description="TM-score=%f" % tm_score),
            SeqRecord(Seq(ref_seq),
                      id=ref_pdb_id,
                      description="TM-score=%f" % tm_score)), tm_score


def extract_chain_from_sup(sup_file,
                           extract_chain,
                           output_file,
                           new_chain_id=None,
                           QUIET=True):
    """ Extract a chain from the superposition file generated by TM-align.

        .. warning::
            TM-align modifies the id of the original chains, so it is highly
            recommended to rename it to match the original chain id.
            To do so, use the parameter ``new_chain_id``.

        Parameters
        ----------
        sup_file: str
            A superposition file generated by TM-align.
        extract_chain: {'A', 'B'}
            Target chain id to be extracted. TM-align performs pair-wise
            structural alignment, so the output file will always contain
            two chains 'A' and 'B', where 'A' and 'B' are the aligned and
            reference structures, respectively.
        output_file: str
            Save the extracted chain to this file.
        new_chain_id: str, optional
            The new chain id of the extracted chain.
            If not provided, the chain ids will be maintained as it is in the
            TM-align output.
        QUIET: bool
            If True (the default), mute warning messages generated by
            Biopython.
    """
    if QUIET:
        try:
            warnings.filterwarnings("ignore")
            logger.debug("Quiet mode activated. From now on, no warning will "
                         "be printed.")
        except Exception:
            logger.warning("Quiet mode could not be activated.")

    if extract_chain not in ["A", "B"]:
        raise ValueError("Valid values for 'extract_chain' are 'A' and 'B'.")

    try:
        logger.debug("Trying to parse the file '%s'." % sup_file)

        structure = parse_from_file("SUP", sup_file)

        model = structure[0]
        if (len(model.child_list) != 2):
            error_msg = ("This structure has %d chains. The file generated "
                         "by TM-align must have 2 chains."
                         % len(model.child_list))
            raise InvalidSuperpositionFileError(error_msg)

        chain_to_remove = 'B' if extract_chain == "A" else "A"
        model.detach_child(chain_to_remove)

        if new_chain_id is not None and extract_chain != new_chain_id:
            model[extract_chain].id = new_chain_id
            logger.debug("Modifications completed.")

        save_to_file(structure, output_file)

        logger.debug("File '%s' created successfully." % output_file)
    except Exception as e:
        logger.exception(e)
        raise


def remove_sup_files(path):
    """ Remove all superposition files created by TM-align at a defined
    directory ``path``."""
    try:
        if (is_directory_valid(path)):
            targets = ['*.sup', '*.sup_atm', '*.sup_all',
                       '*.sup_all_atm', '*.sup_all_atm_lig']

            for target in targets:
                files = glob.glob('%s/%s' % (path, target))
                for file in files:
                    try:
                        remove(file)
                    except Exception as e:
                        logger.exception(e)
                        logger.warning("File %s not removed." % file)
    except Exception as e:
        logger.exception(e)
        raise



from subprocess import Popen, PIPE

from luna.util.exceptions import (InvalidFileFormat, IllegalArgumentError,
                                  ProcessingFailed)
from luna.util.default_values import ANTECHAMBER
from luna.util.file import is_file_valid, get_file_format

import logging
logger = logging.getLogger()


INFORMATS = ["ac", "mol2", "pdb", "mpdb", "prepi", "prepc", "gzmat",
             "gcrt", "mopint", "mopcrt", "gout", "mopout", "alc",
             "csd", "mdl", "hin", "rst"]
OUTFORMATS = INFORMATS

CHARGE_METHODS = ["resp", "cm2", "mul", "rc", "bcc", "esp", "gas", "wc"]

ATOM_TYPES = ["gaff", "amber", "bcc", "sybyl"]


def _prep_opts(opts, prefix=""):
    opt_list = []
    if opts is not None:
        for key in opts:
            opt_list.append("-%s%s" % (prefix, key))

            if opts[key] is not None:
                opt_list.append(str(opts[key]))
    return opt_list


def add_charges(input_file, output_file, total_charge, input_format=None,
                output_format=None, charge_method="bcc", atom_type="sybyl",
                opts=None, antechamber=ANTECHAMBER):

    logger.debug("Charges will be added to the molecule at '%s'." % input_file)

    if is_file_valid(input_file):
        if input_format is None:
            logger.debug("Input file format not defined. "
                         "It will assume the format from the file extension.")
            input_format = get_file_format(input_file)

        if input_format not in INFORMATS:
            raise InvalidFileFormat("Input format '%s' is not accepted "
                                    "by Antechamber." % input_format)

        logger.debug("Input format: %s" % input_format)
    else:
        raise OSError("File '%s' does not exist or is not a valid file."
                      % input_format)

    if isinstance(output_file, str):
        if output_format is None:
            logger.debug("Output file format not defined. "
                         "It will assume the format by the file extension.")
            output_format = get_file_format(output_file, 1)

        if output_format not in OUTFORMATS:
            raise InvalidFileFormat("Input format '%s' does not exist."
                                    % output_format)

        logger.debug("Output format: %s" % output_format)
    else:
        raise IllegalArgumentError("'output_file' must be a string.")

    if charge_method not in CHARGE_METHODS:
        raise IllegalArgumentError("The charge method '%s' does not exist "
                                   "in Antechamber." % charge_method)

    if atom_type not in ATOM_TYPES:
        raise IllegalArgumentError("The atom type '%s' does not exist "
                                   "in Antechamber." % atom_type)

    opts = opts or {}
    if "nc" in opts:
        logger.warning("The parameter for the total charge found "
                       "in 'opts' will be overwritten.")
    if "c" in opts:
        logger.warning("The parameter for the charge method found "
                       "in 'opts' will be overwritten.")
    if "at" in opts:
        logger.warning("The parameter for the atom type found "
                       "in 'opts' will be overwritten.")

    opts["nc"] = total_charge
    opts["c"] = charge_method
    opts["at"] = atom_type
    opts["pf"] = "y"
    opt_list = _prep_opts(opts)

    input_list = ['-i', input_file, "-fi", input_format]
    output_list = ['-o', output_file, "-fo", output_format]

    args = [antechamber] + input_list + output_list + opt_list

    p = Popen(args, stdout=PIPE, stderr=PIPE)
    try:
        stdout, stderr = p.communicate()
    except Exception:
        p.kill()
        raise

    error_lines = stderr.decode().strip().split("\n")

    if len(error_lines) == 1 and error_lines[0] == "":
        logger.debug("File '%s' created with success." % output_file)
    else:
        raise ProcessingFailed("Antechamber could not add charges "
                               "to the provided molecule.")



from operator import xor
from enum import Enum, auto
from collections import defaultdict, Counter

from Bio.PDB.Polypeptide import is_aa

from luna.mol.precomp_data import DefaultResidueData
from luna.wrappers.base import MolWrapper, BondType, OBBondType
from luna.pdb.neighbors.residue import get_residue_neighbors

from openbabel.openbabel import OBSmartsPattern

import logging

logger = logging.getLogger()


METALS = ["Li", "Na", "K", "Rb", "Cs", "Fr", "Be", "Mg", "Ca", "Sr", "Ba",
          "Ra", "Sc", "Ti", "V", "Cr", "Mn", "Fe", "Co", "Ni", "Cu", "Zn",
          "Al", "Ga", "Y", "Zr", "Nb", "Mo", "Tc", "Ru", "Rh", "Pd", "Ag",
          "Cd", "In", "Sn", "Hf", "Ta", "W", "Re", "Os", "Ir", "Pt", "Au",
          "Hg", "Tl", "Pb", "Bi"]
METAL_ATOM = "[%s]" % ",".join(METALS)


class ResidueStandard(Enum):
    """An enumeration of protonation states for standard residues."""

    # Aspartate with double bond on OD1.
    ASP = auto()
    # Same as ASP.
    ASP_OD1 = auto()
    # Aspartate with double bond on OD2.
    ASP_OD2 = auto()

    # Glutamate with double bond on OE1.
    GLU = auto()
    # Same as GLU.
    GLU_OE1 = auto()
    # Glutamate with double bond on OE2.
    GLU_OE2 = auto()

    # HIS: same as HIE.
    HIS = auto()
    # HID as in Amber: histidine with hydrogen on the delta nitrogen.
    HID = auto()
    # HIE as in Amber: histidine with hydrogen on the epsilon nitrogen.
    HIE = auto()
    # HIP as in Amber: histidine with hydrogens on both nitrogens, i.e.,
    # this is positively charged.
    HIP = auto()
    # Negatively charged HIS: both ND1 and NE2 are deprotonated.
    # This is usually used for HIS coordinating two metals.
    HIS_D = auto()

    # CYS as in Amber: neutral cysteine
    CYS = auto()
    # CYM as in Amber: deprotonated cysteine.
    CYM = auto()

    # LYS as in Amber: protonated lysine
    LYS = auto()
    # LYN as in Amber: neutral lysine.
    LYN = auto()

    # Default SER
    SER = auto()
    # SER with the hydroxyl group deprotonated.
    SER_D = auto()

    # Default THR
    THR = auto()
    # THR with the hydroxyl group deprotonated.
    THR_D = auto()

    # Default TYR
    TYR = auto()
    # TYR with the hydroxyl group deprotonated.
    TYR_D = auto()


class Standardizer:

    def __init__(self,
                 asp_type=ResidueStandard.ASP_OD1,
                 cys_type=ResidueStandard.CYS,
                 glu_type=ResidueStandard.GLU_OE1,
                 his_type=ResidueStandard.HIE,
                 lys_type=ResidueStandard.LYS,
                 res_data=DefaultResidueData()):

        self.asp_type = asp_type
        self.cys_type = cys_type
        self.glu_type = glu_type
        self.his_type = his_type
        self.lys_type = lys_type

        self.res_data = res_data

    def _validate_atoms_list(self, res, atom_names, precomp_data=None):
        # Current list of atoms found in this residue.
        atom_names = set(atom_names)

        precomp_data = precomp_data or self.res_data

        # Set of expected atom names.
        data = precomp_data.get(res.resname, {})
        props = data.get("atoms", {})
        expected_atms = set([a for a in props if "," not in a])

        if len(expected_atms) == 0:
            return

        # Identify missing atoms, except OXT.
        missing_atoms = [a for a in expected_atms - atom_names
                         if a != "OXT"]

        # Warn about the missing atoms.
        if missing_atoms:
            logger.warning("The following atoms were not found for "
                           "residue %s: %s."
                           % (res.hierarchy_name,
                              ", ".join(sorted(missing_atoms))))

        # Unexpected atoms found in this residue.
        other_atoms = [a for a in atom_names - expected_atms]

        # Warn about the unexpected atom names identified.
        if other_atoms:
            logger.warning("The following unexpected atoms were found "
                           "for residue %s: %s."
                           % (res.hierarchy_name,
                              ", ".join(sorted(other_atoms))))

    def _resolve_resname(self, res):

        atom_pairs = self._comps[res]

        # Check if ASP is bound to metals.
        if res.resname == "ASP":
            try:
                OD2_atm_obj = [atm_obj for atm_obj, pdb_atm in atom_pairs
                               if pdb_atm.name == "OD2"][0]
            except IndexError:
                # In case OD2 cannot be recovered, return the
                # ASP type defined by the user.
                return self.asp_type.name

            # OD1 -- M.
            smarts1 = f"[OX1]=[CX3]([#6])[OX2;$(O{METAL_ATOM})]"
            # OD2 -- M.
            smarts2 = f"[OX2]({METAL_ATOM})[CX3]([#6])=[OX1]"

            coords = self.metals_coord.get(res, {})
            O_atms = [a.name for a in coords if a.name in ["OD1", "OD2"]]

            # If OD1 is bound to a metal, move the double bond to OD2.
            if (OD2_atm_obj.matches_smarts(smarts1)
                    or ("OD1" in O_atms and "OD2" not in O_atms)):
                return "ASP_OD2"
            # If OD2 is bound to a metal, move the double bond to OD1.
            elif (OD2_atm_obj.matches_smarts(smarts2)
                    or ("OD1" not in O_atms and "OD2" in O_atms)):
                return "ASP_OD1"
            # Otherwise, return the ASP type defined by the user.
            return self.asp_type.name

        # Check if CYS is bound to metals.
        elif res.resname == "CYS":
            try:
                SG_atm_obj = [atm_obj for atm_obj, pdb_atm in atom_pairs
                              if pdb_atm.name == "SG"][0]
            except IndexError:
                # In case SG cannot be recovered, return the
                # CYS type defined by the user.
                return self.cys_type.name

            coords = self.metals_coord.get(res, {})
            trgt_atms = [a.name for a in coords if a.name in ["SG"]]

            # If SG is bound to a metal, deprotonate it.
            if (SG_atm_obj.matches_smarts(f"[S]{METAL_ATOM}")
                    or "SG" in trgt_atms):
                return "CYM"
            # Otherwise, return the CYS type defined by the user.
            return self.cys_type.name

        # Check if GLU is bound to metals.
        elif res.resname == "GLU":
            try:
                OE2_atm_obj = [atm_obj for atm_obj, pdb_atm in atom_pairs
                               if pdb_atm.name == "OE2"][0]
            except IndexError:
                # In case OE2 cannot be recovered, return the
                # GLU type defined by the user.
                return self.glu_type.name

            # OE1 -- M.
            smarts1 = f"[OX1]=[CX3]([#6])[OX2;$(O{METAL_ATOM})]"
            # OE2 -- M.
            smarts2 = f"[OX2]({METAL_ATOM})[CX3]([#6])=[OX1]"

            coords = self.metals_coord.get(res, {})
            O_atms = [a.name for a in coords if a.name in ["OE1", "OE2"]]

            # If OE1 is bound to a metal, move the double bond to OE2.
            if (OE2_atm_obj.matches_smarts(smarts1)
                    or "OE1" in O_atms):
                return "GLU_OE2"
            # If OE2 is bound to a metal, move the double bond to OE1.
            elif (OE2_atm_obj.matches_smarts(smarts2)
                    or "OE2" in O_atms):
                return "GLU_OE1"
            # Otherwise, return the GLU type defined by the user.
            return self.glu_type.name

        elif res.resname == "HIS":
            try:
                ND1_atm_obj = [atm_obj for atm_obj, pdb_atm in atom_pairs
                               if pdb_atm.name == "ND1"][0]
            except IndexError:
                # In case ND1 cannot be recovered, return the
                # HIS type defined by the user.
                return self.his_type.name

            # HIS bound to two different metals:  M1 <--- ND1 ... NE2 ---> M2.
            smarts1 = f"[#7;R]({METAL_ATOM})~[#6;R]~[#7;R]{METAL_ATOM}"
            # ND1 ---> M1
            smarts2 = f"[#7;R]{METAL_ATOM}"
            # NE2 ---> M1
            smarts3 = f"[#7;R]~[#6;R]~[#7;R]{METAL_ATOM}"

            coords = self.metals_coord.get(res, {})
            N_atms = [a.name for a in coords if a.name in ["ND1", "NE2"]]

            # If both Ns are bound to the metal, deprotonate the HIS.
            if (ND1_atm_obj.matches_smarts(smarts1)
                    or len(N_atms) == 2):
                return "HIS_D"
            # If ND1 is bound to the metal, the H is added to NE2.
            elif (ND1_atm_obj.matches_smarts(smarts2)
                    or "ND1" in N_atms):
                return "HIE"
            # If NE2 is bound to the metal, the H is added to ND1.
            elif (ND1_atm_obj.matches_smarts(smarts3)
                    or "NE2" in N_atms):
                return "HID"
            # Otherwise, return the HIS type defined by the user.
            return self.his_type.name

        elif res.resname == "LYS":
            try:
                NZ_atm_obj = [atm_obj for atm_obj, pdb_atm in atom_pairs
                              if pdb_atm.name == "NZ"][0]
            except IndexError:
                # In case NZ cannot be recovered, return the
                # LYS type defined by the user.
                return self.lys_type.name

            coords = self.metals_coord.get(res, {})
            trgt_atms = [a.name for a in coords if a.name in ["NZ"]]

            # If NZ is bound to a metal, deprotonate it.
            if (NZ_atm_obj.matches_smarts(f"[N]{METAL_ATOM}")
                    or "NZ" in trgt_atms):
                return "LYN"
            # Otherwise, return the LYS type defined by the user.
            return self.lys_type.name

        elif res.resname == "TYR":
            try:
                OH_atm_obj = [atm_obj for atm_obj, pdb_atm in atom_pairs
                              if pdb_atm.name == "OH"][0]
            except IndexError:
                # In case OG/OG1/OH cannot be recovered, return the
                # residue type defined by the user.
                return res.resname

            coords = self.metals_coord.get(res, {})
            trgt_atms = [a.name for a in coords
                         if a.name == "OH"]

            # If the oxygen is bound to a metal, deprotonate it.
            if (OH_atm_obj.matches_smarts(f"[O]{METAL_ATOM}")
                    or "OH" in trgt_atms):
                return res.resname + "_D"
            # Otherwise, return the residue type defined by the user.
            return res.resname

        # If any rule applies to this residue, return its current name.
        return res.resname

    def _check_metal_coordination(self, atm_obj, pdb_atm):

        res = pdb_atm.parent
        n_heavy_atoms = \
            atm_obj.get_neighbors_number(only_heavy_atoms=True)

        # If this residue is not bound to any atom, then its valence
        # is correct.
        if n_heavy_atoms == 0:
            return res.resname

        # Metal properties.
        data = self.res_data.get(res.resname, {})
        props = data.get("atoms", {})

        # Coordination number.
        coords = props[pdb_atm.name].get("coords", [])
        coords = [int(i) for i in coords.split(",")]

        # Evaluate if the coordination number is correct.
        if n_heavy_atoms not in coords:
            msg = ("An unexpected coordination number (%d) has been "
                   "found for metal %s.")
            logger.warning(msg % (n_heavy_atoms, pdb_atm.hierarchy_name))

    def _get_partner_atm(self, atm_obj, pdb_atm, bond_obj):
        partner_obj = \
            bond_obj.get_partner_atom(atm_obj)
        partner_idx = partner_obj.get_idx()

        # If this atom's partner cannot be identified, it is better
        # not to update this atom.
        if partner_idx not in self._pdb_mapping:
            self._events["ignore"].add(pdb_atm)

            msg = ("The atom %s from %s will not be amended "
                   "because an unidentified atom is covalently bound "
                   "to it." % (pdb_atm.name, pdb_atm.parent.hierarchy_name))
            logger.warning(msg)

            return None, None

        partner_atm = self._pdb_mapping[partner_idx]

        return partner_obj, partner_atm

    def _resolve_bonds(self, resname, atm_obj, pdb_atm, precomp_data=None):

        precomp_data = precomp_data or self.res_data

        # All expected residue bonds.
        res_dict = precomp_data.get(resname, {})
        res_bonds = res_dict.get("bonds", {})

        # Add disulfide bridge as a standard bond to CYS.
        if pdb_atm.parent.resname == "CYS":
            res_bonds["SG,SG"] = {"type": "SINGLE", "is_aromatic": False}

        # Bonds found for atom 'atom_obj', identified by their key.
        found_bonds = set()
        # Set of atom pairs.
        found_atm_pairs = set()

        # Loop through the bonds formed by this atom to check if there
        # is any unexpected bond (e.g., metal bonds).
        for bond_obj in atm_obj.get_bonds():

            # Get this atom's partner.
            partner_obj, partner_atm = \
                self._get_partner_atm(atm_obj, pdb_atm, bond_obj)

            if partner_obj is None or partner_atm is None:
                return

            bond_key = ",".join(sorted([pdb_atm.name, partner_atm.name]))

            # Break any bonds with metals.
            if (atm_obj.get_symbol() in METALS
                    or partner_obj.get_symbol() in METALS):
                self._events["break"].add(((atm_obj, pdb_atm),
                                           (partner_obj, partner_atm),
                                           bond_obj))
                continue

            # Break any bonds with water.
            if pdb_atm.parent.is_water() or partner_atm.parent.is_water():
                self._events["break"].add(((atm_obj, pdb_atm),
                                           (partner_obj, partner_atm),
                                           bond_obj))
                continue

            # If an unexpected bond between residue atoms is found, break it.
            if (is_aa(pdb_atm.parent) and is_aa(partner_atm.parent)
                    and bond_key not in res_bonds):

                # In case there is a peptide bond between a residue and a
                # non-standard residue, whose N has a name different than 'N',
                # we can ignore it if the atom C formed an amide with the N.
                if pdb_atm.name == "C" and partner_obj.get_symbol() == "N":
                    # Correct amide group.
                    if atm_obj.matches_smarts("[$([CX3](=O)N)]"):
                        continue

                    # Fix C-N bond from an amide group incorrectly parsed by
                    # Open Babel.
                    elif atm_obj.matches_smarts("[$([CX3](O)=N)]"):
                        bond_type = BondType["SINGLE"]
                        # Add a new modification event for this bond.
                        self._events["modbonds"].add(((atm_obj,
                                                       pdb_atm),
                                                      (partner_obj,
                                                       partner_atm),
                                                      bond_obj, bond_type,
                                                      False))
                        continue

                self._events["break"].add(((atm_obj, pdb_atm),
                                           (partner_obj, partner_atm),
                                           bond_obj))

                if pdb_atm.parent == partner_atm.parent:
                    msg = ("An unexpected bond between atoms %s and %s from "
                           "%s was found. It will break the bond and "
                           "standardize the atoms."
                           % (pdb_atm, partner_atm, pdb_atm.parent))
                else:
                    msg = ("An unexpected bond between atoms %s and %s from "
                           "residues %s and %s was found. It will break the "
                           "bond and standardize the atoms."
                           % (pdb_atm, partner_atm,
                              pdb_atm.parent, partner_atm.parent))
                logger.warning(msg)
                continue

            # If this atom's partner is not a residue, it is better not to
            # update this atom.
            elif ((is_aa(pdb_atm.parent) and not is_aa(partner_atm.parent))
                    or bond_key not in res_bonds):

                self._events["ignore"].add(pdb_atm)

                msg = ("The atom %s from %s will not be amended "
                       "because an unexpected atom (%s) from %s is covalently "
                       "bound to it." % (pdb_atm.name,
                                         pdb_atm.parent.hierarchy_name,
                                         partner_atm.name,
                                         partner_atm.parent.hierarchy_name))
                logger.warning(msg)
                return

            # Update the sets with the new identified pair.
            found_bonds.add(bond_key)
            found_atm_pairs.add(tuple(sorted([pdb_atm, partner_atm])))

            # Get bond type from the file definition.
            bond_type = BondType[res_bonds[bond_key]["type"]]
            # Get bond aromaticity from the file definition.
            is_aromatic = res_bonds[bond_key]["is_aromatic"]

            # If the current bond is as expected, don't modify it.
            if (bond_type == bond_obj.get_bond_type()
                    and is_aromatic == bond_obj.is_aromatic()):
                continue

            # Add a new modification event for this bond.
            self._events["modbonds"].add(((atm_obj, pdb_atm),
                                          (partner_obj, partner_atm),
                                          bond_obj, bond_type,
                                          is_aromatic))

        # Set of expected bonds.
        expected_bonds = set([b for b in res_bonds
                              for an in b.split(",") if pdb_atm.name == an])

        # Set of missing bonds.
        # It ignores missing disulfide bonds as they may not exist.
        missing_bonds = [b for b in expected_bonds - set(found_bonds)
                         if b != "SG,SG"]

        # Only verify missing bonds for residues.
        if missing_bonds and pdb_atm.parent.is_residue():
            nb_residues = get_residue_neighbors(pdb_atm.parent, verbose=False)

            # Loop through each missing bond and add them in case
            # both atoms exist.
            for bond_key in missing_bonds:
                bond_pair = bond_key.split(",")

                partner_atm_name = (bond_pair[1]
                                    if bond_pair[0] == pdb_atm.name
                                    else bond_pair[0])

                partner_res = pdb_atm.parent
                if pdb_atm.name == "N" and partner_atm_name == "C":
                    # There's nothing to be done because the previous
                    # residue was not selected or is missing.
                    if not nb_residues.has_predecessor():
                        continue
                    partner_res = nb_residues.predecessor

                elif pdb_atm.name == "C" and partner_atm_name == "N":
                    # There's nothing to be done because the next
                    # residue was not selected or is missing.
                    if not nb_residues.has_successor():
                        continue
                    partner_res = nb_residues.successor

                trg_atm_pairs = []
                # If the partner residue exists, search for the partner
                # atom by name.
                if partner_res in self._comps:
                    atom_pairs = self._comps[partner_res]
                    trg_atm_pairs = [(o, p) for o, p in atom_pairs
                                     if (p.parent == partner_res
                                         and p.name == partner_atm_name)]

                # Skip missing bonds related to C- and N-terminal residues.
                if (bond_key in ["C,N", "C,OXT"]
                        and len(trg_atm_pairs) == 0):
                    continue

                # If the partner atom could not be identified, skip it.
                elif len(trg_atm_pairs) == 0:
                    self._events["ignore"].add(pdb_atm)

                    msg = ("The bond '%s' was not found for residue %s. "
                           "However, it cannot be added because the atom %s "
                           "seems not to exist."
                           % (bond_key, pdb_atm.parent.hierarchy_name,
                              partner_atm_name))
                    logger.warning(msg)
                    continue

                # Partner atom's object
                partner_obj, partner_atm = trg_atm_pairs[0]
                # Bond type
                bond_name = res_bonds[bond_key]["type"]
                bond_type = OBBondType[bond_name]
                # Bond aromaticity
                is_aromatic = res_bonds[bond_key]["is_aromatic"]

                # Create a new bond modification event.
                self._events["modbonds"].add(((atm_obj, pdb_atm),
                                              (partner_obj, partner_atm),
                                              None, bond_type,
                                              is_aromatic))

    def _resolve_carbox_invariants(self, trgt_atm_obj, trgt_pdb_atm):

        smarts1 = "[OX1]=[CX3][OX1H0-,OX2H1]"
        smarts2 = "[OX1H0-,OX2H1][CX3]=[OX1]"
        if (not trgt_atm_obj.matches_smarts(smarts1)
                and not trgt_atm_obj.matches_smarts(smarts2)):
            return

        bond_obj = trgt_atm_obj.get_bonds()[0]

        C_atm_obj = bond_obj.get_partner_atom(trgt_atm_obj)

        other_O_atm_obj = None
        other_O_pdb_atm = None

        for C_bond_obj in C_atm_obj.get_bonds():
            C_partner_obj = \
                C_bond_obj.get_partner_atom(C_atm_obj)

            if (C_partner_obj.get_symbol() == "O"
                    and C_partner_obj.get_id() != trgt_atm_obj.get_id()):

                other_O_atm_obj = C_partner_obj
                other_O_pdb_atm = self._pdb_mapping[C_partner_obj.get_idx()]

        # If atomic data from the other carbox O were not found, return None.
        if other_O_atm_obj is None or other_O_pdb_atm is None:
            return

        coords = self.metals_coord.get(trgt_pdb_atm.parent, {})

        # M1 <--- O-C=O ---> M2
        #
        #   if both Os coordinate a metal, it's not necessary to
        #   update their invariants.
        #
        if trgt_pdb_atm in coords and other_O_pdb_atm in coords:
            return

        #     V
        # O-C=O ---> M
        #
        #   if the O with the double bond coordinates a metal,
        #   update its invariant.
        if (trgt_atm_obj.matches_smarts(smarts1)
                and trgt_pdb_atm in coords):

            # Fix the bond: double to single bond.
            C_pdb_atm = self._pdb_mapping[C_atm_obj.get_idx()]
            self._events["modbonds"].add(((trgt_atm_obj, trgt_pdb_atm),
                                          (C_atm_obj, C_pdb_atm),
                                          bond_obj, BondType["SINGLE"],
                                          False))
            return [1, 1, 8, 16, -1, 0, 0]

        # V
        # O-C=O ---> M
        #
        #   if the O with the single bond has a partner O coordinating
        #   a metal, update its invariant.
        elif (trgt_atm_obj.matches_smarts(smarts2)
                and other_O_pdb_atm in coords):

            # Fix the bond: single to double bond.
            C_pdb_atm = self._pdb_mapping[C_atm_obj.get_idx()]
            self._events["modbonds"].add(((trgt_atm_obj, trgt_pdb_atm),
                                          (C_atm_obj, C_pdb_atm),
                                          bond_obj, BondType["DOUBLE"],
                                          False))
            return [1, 2, 8, 16, 0, 0, 0]

    def _resolve_imidazole_invariants(self, trgt_atm_obj, trgt_pdb_atm,
                                      precomp_data=None):

        # Imidazole-like, but not tetrazoles.
        strict_smarts_rule = ("[$(nc[n;H1]),$([n;H1]cn);R1r5;"
                              "!$([nR1r5;$(n:n:n:n:c),$(n:n:n:c:n)])]")

        generic_smarts_rule = ("[$([#7]~[#6]~[#7]);R1r5;"
                               "!$([nR1r5;$(n:n:n:n:c),$(n:n:n:c:n)])]")

        if not trgt_atm_obj.matches_smarts(generic_smarts_rule):
            return

        C_atm_obj = None
        C_bond_obj = None
        for bond_obj in trgt_atm_obj.get_bonds():
            partner_obj = bond_obj.get_partner_atom(trgt_atm_obj)

            if partner_obj.matches_smarts("[$([#6](~[#7])~[#7])]"):
                C_atm_obj = partner_obj
                C_bond_obj = bond_obj
                break

        other_N_pdb_atm = None

        for bond_obj in C_atm_obj.get_bonds():
            C_partner_obj = \
                bond_obj.get_partner_atom(C_atm_obj)

            if (C_partner_obj.get_symbol() == "N"
                    and C_partner_obj.get_id() != trgt_atm_obj.get_id()):

                other_N_pdb_atm = self._pdb_mapping[C_partner_obj.get_idx()]

        is_N_single = False
        is_N_double = False
        if trgt_atm_obj.matches_smarts(strict_smarts_rule):
            if trgt_atm_obj.matches_smarts("[n+0X3H1]"):
                is_N_single = True
            elif trgt_atm_obj.matches_smarts("[n+1X3H1,n+0X2H0]"):
                is_N_double = True

        elif precomp_data is not None:
            res_dict = precomp_data.get(trgt_pdb_atm.parent.resname, {})
            bonds_count = Counter([res_dict["bonds"][b]["type"]
                                   for b in res_dict["bonds"]
                                   for a in b.split(",")
                                   if a == trgt_pdb_atm.name])

            if bonds_count["SINGLE"] == 2:
                is_N_single = True
            elif bonds_count["SINGLE"] == 1 and bonds_count["DOUBLE"] == 1:
                is_N_double = True

        # Both False or both True
        if is_N_single == is_N_double:
            return

        coords = self.metals_coord.get(trgt_pdb_atm.parent, {})

        # M1 <--- n1 ... n2 ---> M2.
        #
        #   if both Ns coordinate a metal, remove the Hs from both Ns.
        if trgt_pdb_atm in coords and other_N_pdb_atm in coords:
            if is_N_single:
                return [2, 2, 7, 14, -1, 0, 1]
            return [2, 3, 7, 14, 0, 0, 1]

        #       v
        #  n1-c=n2 ---> M1.
        #
        #   if the N with the double bond coordinates a metal and have a
        #       H, fix its charge.
        if trgt_pdb_atm in coords and is_N_double:
            return [2, 3, 7, 14, 0, 0, 1]

        #  v
        #  n1-c=n2 ---> M1.
        #
        #   if the N with the single bond has a partner N coordinating
        #       a metal, fix it if necessary.
        if other_N_pdb_atm in coords and is_N_single:
            return [2, 2, 7, 14, 0, 1, 1]

        #       v
        #  n2=c-n1 ---> M1.
        #
        #   if the N with the single bond coordinates a metal, move the
        # double bond to it.
        if trgt_pdb_atm in coords and is_N_single:
            # Fix the bond: single to double bond.
            C_pdb_atm = self._pdb_mapping[C_atm_obj.get_idx()]
            self._events["modbonds"].add(((trgt_atm_obj, trgt_pdb_atm),
                                          (C_atm_obj, C_pdb_atm),
                                          C_bond_obj, BondType["DOUBLE"],
                                          True))
            return [2, 3, 7, 14, 0, 0, 1]

        #  v
        #  n2=c-n1 ---> M1.
        #
        #   if the N with the double bond has a partner N coordinating
        #       a metal, move the single bond to it.
        elif other_N_pdb_atm in coords and is_N_double:

            # Fix the bond: double to single bond.
            C_pdb_atm = self._pdb_mapping[C_atm_obj.get_idx()]
            self._events["modbonds"].add(((trgt_atm_obj, trgt_pdb_atm),
                                          (C_atm_obj, C_pdb_atm),
                                          C_bond_obj, BondType["SINGLE"],
                                          True))
            return [2, 2, 7, 14, 0, 1, 1]

    def _resolve_tetrazole_invariants(self, trgt_atm_obj, trgt_pdb_atm,
                                      precomp_data=None):

        precomp_data = precomp_data or {}

        # Tetrazole-like.
        strict_smarts_rule = ("[nR1r5;$(n:n:n:n:c),$(n:n:n:c:n)]")
        generic_smarts_rule = ("[#7R1r5;$([#7]~[#7]~[#7]~[#7]~[#6]),"
                               "$([#7]~[#7]~[#7]~[#6]~[#7])]")

        if not trgt_atm_obj.matches_smarts(generic_smarts_rule):
            return

        # The next lines will verify if the current N contains
        # a double bond or only single bonds.
        is_N_single = False
        is_N_double = False
        if trgt_atm_obj.matches_smarts(strict_smarts_rule):
            if trgt_atm_obj.matches_smarts("[n;X3H1+0,X2H0-1]"):
                is_N_single = True
            elif trgt_atm_obj.matches_smarts("[nX2H0+0]"):
                is_N_double = True
        elif precomp_data is not None:
            res_dict = precomp_data.get(trgt_pdb_atm.parent.resname, {})
            bonds_count = Counter([res_dict["bonds"][b]["type"]
                                   for b in res_dict["bonds"]
                                   for a in b.split(",")
                                   if a == trgt_pdb_atm.name])

            if bonds_count["SINGLE"] == 2:
                is_N_single = True
            elif bonds_count["SINGLE"] == 1 and bonds_count["DOUBLE"] == 1:
                is_N_double = True

        # Error: return if both False or both True.
        if is_N_single == is_N_double:
            return

        # In the next line, it will capture the tetrazole group that contains
        # the current atom.
        ob_smart = OBSmartsPattern()
        ob_smart.Init("[#6]1~[#7]~[#7]~[#7]~[#7H,#7-1]1")
        ob_smart.Match(trgt_atm_obj.parent.unwrap())

        atoms = []
        matches = [x for x in ob_smart.GetMapList()]
        if matches:
            for match in matches:
                trgt_grp_found = False
                for idx in match:
                    pdb_atm = self._pdb_mapping[idx]
                    atoms.append(pdb_atm)

                    if pdb_atm == trgt_pdb_atm:
                        trgt_grp_found = True

                if trgt_grp_found:
                    break

        # Error: if no atom group has been found by the previous
        # SMARTS definition.
        if not atoms:
            return

        # Count how many metals this tetrazole coordinates.
        n_coords = len([a for a in atoms if a.has_metal_coordination()])

        # If the tetrazole coordinates only one metal, move the negative
        # charge to the N coordinating the metal.
        if n_coords == 1:
            # If the current atom is the one coordinating the metal.
            if trgt_pdb_atm.has_metal_coordination():

                # If the current atom contains a double bond, it will be
                # necessary to standardize all bonds in the tetrazole.
                if is_N_double:
                    # Atoms in the opposite end to the current atom.
                    opp_atm_obj, opp_pdb_atm = None, None
                    # Set of imediate neighbors of the current atom.
                    first_bonded_atms = set()
                    for bond_obj1 in trgt_atm_obj.get_bonds():
                        partner_obj1 = \
                            bond_obj1.get_partner_atom(trgt_atm_obj)
                        pdb_atm1 = self._pdb_mapping[partner_obj1.get_idx()]

                        mod_bond = ((trgt_atm_obj, trgt_pdb_atm),
                                    (partner_obj1, pdb_atm1),
                                    bond_obj1,
                                    BondType["SINGLE"], True)
                        self._events["modbonds"].add(mod_bond)

                        first_bonded_atms.add(pdb_atm1)

                        for bond_obj2 in partner_obj1.get_bonds():
                            partner_obj2 = \
                                bond_obj2.get_partner_atom(partner_obj1)
                            pdb_atm2 = \
                                self._pdb_mapping[partner_obj2.get_idx()]

                            if pdb_atm2 == trgt_pdb_atm:
                                continue
                            if pdb_atm2 not in atoms:
                                continue

                            if opp_atm_obj is None:
                                opp_atm_obj = partner_obj2
                                opp_pdb_atm = pdb_atm2

                            mod_bond = ((partner_obj1, pdb_atm1),
                                        (partner_obj2, pdb_atm2),
                                        bond_obj2,
                                        BondType["DOUBLE"], True)
                            self._events["modbonds"].add(mod_bond)

                    # Standardize the bonds in the opposite end.
                    for bond_obj in opp_atm_obj.get_bonds():
                        partner_obj = \
                            bond_obj.get_partner_atom(opp_atm_obj)
                        pdb_atm = self._pdb_mapping[partner_obj.get_idx()]

                        if pdb_atm in first_bonded_atms:
                            continue

                        mod_bond = ((opp_atm_obj, opp_pdb_atm),
                                    (partner_obj, pdb_atm),
                                    bond_obj,
                                    BondType["SINGLE"], True)
                        self._events["modbonds"].add(mod_bond)
                return [2, 2, 7, 14, -1, 0, 1]
            return [2, 3, 7, 14, 0, 0, 1]

        # If the tetrazole coordinates more than 1 metal, standardize it
        # according to the PDB.
        else:
            self._resolve_bonds(trgt_pdb_atm.parent.resname,
                                trgt_atm_obj,
                                trgt_pdb_atm,
                                precomp_data)

            if is_N_double:
                return [2, 3, 7, 14, 0, 0, 1]
            return [2, 2, 7, 14, -1, 0, 1]

    def _resolve_sulfur_groups_invariants(self, trgt_atm_obj, trgt_pdb_atm,
                                          precomp_data=None):

        precomp_data = precomp_data or {}
        res_dict = precomp_data.get(trgt_pdb_atm.parent.resname, {})
        res_bonds = res_dict.get("bonds", {})

        if not res_bonds:
            return

        bad_sulf_acid_O_smarts = "[$([OH1]S([OH1])([OH1]))]"
        if not trgt_atm_obj.matches_smarts(bad_sulf_acid_O_smarts):
            return

        bond_obj = trgt_atm_obj.get_bonds()[0]

        S_atm_obj = bond_obj.get_partner_atom(trgt_atm_obj)
        S_pdb_atm = self._pdb_mapping[S_atm_obj.get_idx()]

        bond_key = ",".join(sorted([trgt_pdb_atm.name,
                                    S_pdb_atm.name]))
        bond_dict = res_bonds[bond_key]

        if bond_dict["type"] == "SINGLE":
            return [1, 1, 8, 16, -1, 0, 0]

        if bond_dict["type"] == "DOUBLE":
            return [1, 2, 8, 16, 0, 0, 0]

    def _resolve_invariants(self, resname, atm_obj, pdb_atm,
                            atm_names=None, precomp_data=None):

        if pdb_atm in self._events["ignore"]:
            return

        res = pdb_atm.parent
        atm_names = atm_names or []
        precomp_data = precomp_data or self.res_data

        # Expected invariant.
        invariants = None
        # Define if this atom is aromatic.
        is_aromatic = None

        if res.is_hetatm() or res.is_nucleotide():
            coords = self.metals_coord.get(res, {})

            res_dict = precomp_data.get(resname, {})
            res_props = res_dict.get("props", {})

            # Check if the atom aromaticity is incorrect.
            fix_atm_aromaticity = False
            if (pdb_atm.name in res_props
                    and "AROMATIC" in res_props[pdb_atm.name]["features"]
                    and not atm_obj.is_aromatic()):
                fix_atm_aromaticity = True

            # Currently, only standardization rules for O, N, and S were
            # defined. For that reason, only these atoms pass the following
            # condition or atoms whose aromaticity was incorrectly perceived.
            if (atm_obj.get_symbol() not in ["O", "N", "S"]
                    and not fix_atm_aromaticity):
                return

            # Fix O atoms
            if atm_obj.get_symbol() == "O":

                # Carbonate groups.
                carbonate_smarts = \
                    ("[$([OX1]=[CX3]([OX1H0-,OX2H1])[OX1H0-,OX2H1])"
                     ",$([OX1H0-,OX2H1][CX3]([OX1H0-,OX2H1])=[OX1])]")
                # Carboxylic acid, but not carbonate.
                carbox_smarts = ("[$([OX1]=[CX3][OX1H0-,OX2H1])"
                                 ",$([OX1H0-,OX2H1][CX3]=[OX1])"
                                 ";!$(%s)]" % carbonate_smarts)
                phosphate_smarts = "[$([OX2H1]P([OX2H1])([OX2H1])=[OX1])]"
                # O from sulfuric or sulfonic acid perceived as OH.
                bad_sulf_acid_O_smarts = "[$([OH1]S([OH1])([OH1]))]"

                # Carboxylic group O coordinating a metal.
                if atm_obj.matches_smarts(carbox_smarts):
                    invariants = \
                        self._resolve_carbox_invariants(atm_obj, pdb_atm)

                # Force all OH from phosphates to be deprotonated.
                elif atm_obj.matches_smarts(phosphate_smarts):
                    invariants = [1, 1, 8, 16, -1, 0, 0]

                # Phosphoric, phosphonic, or phosphinic acid O
                # coordinating a metal.
                elif (atm_obj.matches_smarts("[$([OX2H1]P(=[OX1]))]")
                        and pdb_atm in coords):
                    invariants = [1, 1, 8, 16, -1, 0, 0]

                # Sulfuric acid O perceived as OH.
                elif atm_obj.matches_smarts(bad_sulf_acid_O_smarts):
                    invariants = \
                        self._resolve_sulfur_groups_invariants(atm_obj,
                                                               pdb_atm,
                                                               precomp_data)
                # Phenol O coordinating a metal as in TYR.
                elif (atm_obj.matches_smarts("[OH+0]c")
                        and pdb_atm in coords):
                    invariants = [1, 1, 8, 16, -1, 0, 0]

            # Fix N atoms.
            elif atm_obj.get_symbol() == "N":

                # Amines.
                amine_smarts = "[$([NX4+;H3,H2,H1;!$(NC=O)][C])]"

                # Tetrazole.
                #
                # This rule doesn't specify aromaticity because some aromatic
                # rings may be perceived a simple cyclic ring by Open Babel.
                generic_tetrazole_N_smarts = \
                    ("[#7R1r5;$([#7]~[#7]~[#7]~[#7]~[#6]),"
                     "$([#7]~[#7]~[#7]~[#6]~[#7])]")

                # Tetrazole.
                #
                # With aromaticity correctly perceived.
                tetrazole_N_smarts = "[nR1r5;$(n:n:n:n:c),$(n:n:n:c:n)]"

                # Imidazole-like, but not tetrazoles.
                #
                # This rule doesn't specify aromaticity because some aromatic
                # rings may be perceived a simple cyclic ring by Open Babel.
                imidazole_smarts = ("[$([#7]~[#6]~[#7]~[#6]~[#6]);R1r5;"
                                    f"!$({generic_tetrazole_N_smarts})]")

                if atm_obj.matches_smarts(generic_tetrazole_N_smarts):
                    if (atm_obj.matches_smarts(tetrazole_N_smarts)
                            or fix_atm_aromaticity):
                        invariants = \
                            self._resolve_tetrazole_invariants(atm_obj,
                                                               pdb_atm,
                                                               precomp_data)
                    if fix_atm_aromaticity:
                        is_aromatic = True

                # Imidazole N coordinating a metal.
                elif atm_obj.matches_smarts(imidazole_smarts):
                    if (atm_obj.matches_smarts('[$(nc[n;H1]),$([n;H1]cn)]')
                            or fix_atm_aromaticity):

                        invariants = \
                            self._resolve_imidazole_invariants(atm_obj,
                                                               pdb_atm,
                                                               precomp_data)

                    if fix_atm_aromaticity:
                        is_aromatic = True

                # Amine N coordinating a metal.
                elif (atm_obj.matches_smarts(amine_smarts)
                        and pdb_atm in coords):
                    invariants = atm_obj.get_atomic_invariants()
                    invariants[4] = 0
                    invariants[5] = invariants[5] - 1

                # Amide N coordinating a metal.
                #
                # Note that only amide groups as in amino acid backbones
                # are considered by this rule.
                elif (atm_obj.matches_smarts("[$([#7;X3H1]C=O)]")
                        and pdb_atm in coords):
                    invariants = [2, 2, 7, 14, -1, 0, 0]

                # Broken amide from residues (including non-standard ones).
                elif (is_aa(pdb_atm.parent)
                        and atm_obj.matches_smarts("[$(N=[CX3]O)]")):
                    invariants = [2, 2, 7, 14, 0, 1, 0]

                # Sulfonamide N coordinating a metal.
                elif (atm_obj.matches_smarts("[$([N;H1,H2]S(=O)(=O))]")
                        and pdb_atm in coords):
                    invariants = atm_obj.get_atomic_invariants()
                    invariants[4] = -1
                    invariants[5] = invariants[5] - 1

            # Fix S atoms
            elif atm_obj.get_symbol() == "S":
                # S from a sulfuric or sulfonic acid where all Os were
                # perceived as OH.
                bad_sulf_acid_S_smarts = "[$(S([OH1])([OH1])([OH1]))]"
                if atm_obj.matches_smarts(bad_sulf_acid_S_smarts):
                    self._resolve_bonds(resname, atm_obj, pdb_atm,
                                        precomp_data)

                # Thiol S coordinating a metal.
                elif (atm_obj.matches_smarts("[SH+0]")
                        and pdb_atm in coords):
                    invariants = [1, 1, 16, 32, -1, 0, 0]

            # Fix C atoms with incorrect aromaticity.
            elif (atm_obj.get_symbol() == "C" and fix_atm_aromaticity
                    and pdb_atm.name in res_props):

                invariants = atm_obj.get_atomic_invariants()

                # Set the correct degree.
                invariants[1] = (res_props[pdb_atm.name]["degree"]
                                 - res_props[pdb_atm.name]["n_Hs"])
                # Set the proper number of Hs.
                invariants[5] = res_props[pdb_atm.name]["n_Hs"]
                # Set the aromaticity flag ON.
                is_aromatic = True

                # Don't update bonds from tetrazole carbons,
                # because their bonds are updated when
                # _resolve_tetrazole_invariants() are called.
                generic_tetrazole_C_smarts = "[$([#6]~[#7]~[#7]~[#7]~[#7])]"
                if not atm_obj.matches_smarts(generic_tetrazole_C_smarts):
                    self._resolve_bonds(resname, atm_obj, pdb_atm,
                                        precomp_data)

        elif res.is_residue():

            if pdb_atm.name in ["C", "N"]:
                nb_residues = \
                    get_residue_neighbors(res, verbose=False)

                coords = self.metals_coord.get(res, {})
                trgt_atms = [a.name for a in coords if a.name in ["N"]]

                if (pdb_atm.name == "N"
                        and nb_residues.predecessor not in self._comps):
                    invariants = [1, 1, 7, 14, 1, 3, 0]
                    if res.resname == "PRO":
                        invariants = [2, 2, 7, 14, 1, 2, 1]

                elif (pdb_atm.name == "N"
                        and atm_obj.matches_smarts(f"[N]{METAL_ATOM}")
                        or "N" in trgt_atms):
                    invariants = [2, 2, 7, 14, -1, 0, 0]

                elif (pdb_atm.name == "C"
                        and nb_residues.successor not in self._comps):

                    n_heavy_atms = \
                        atm_obj.get_neighbors_number(only_heavy_atoms=True)

                    # The following rule accounts for amide bonds between
                    # residues and non-standard residues. That's because the
                    # function get_residue_neighbors() only considers peptide
                    # bonds formed between an atom 'C' and an atom 'N'.
                    # That means the function would ignore a non-standard
                    # residue forming the peptide if its N has a different
                    # name (e.g., N3 as in SNN).
                    amide_like = "[$([CX3](=,-O)=,-N)]"
                    if (nb_residues.successor is None and n_heavy_atms == 3
                            and atm_obj.matches_smarts(amide_like)):
                        invariants = [3, 4, 6, 12, 0, 0, 0]

                    # If it's a C-terminal residue and OXT doesn't exist,
                    # then C has one Hydrogen.
                    elif "OXT" not in atm_names or n_heavy_atms != 3:
                        invariants = [2, 3, 6, 12, 0, 1, 0]

            # CYS:SG in disulfide bridges.
            elif pdb_atm.name == "SG" and atm_obj.matches_smarts("[$(S(C)S)]"):
                invariants = [2, 2, 16, 32, 0, 0, 0]

        # Skip any hetatm and nucleotide if 'invariants' is None.
        #
        # Residues/Water/Metal can pass because their invariants are
        # predefined in the default definition file.
        if (res.is_hetatm() or res.is_nucleotide()) and invariants is None:
            return

        # Load the invariants from the definition file.
        if invariants is None:
            props = precomp_data[resname]["atoms"]
            invariants = props[pdb_atm.name]["invariants"]
            invariants = [int(i) for i in invariants.split(",")]
            is_aromatic = props[pdb_atm.name]["is_aromatic"]

        # Add a new event to modify the atom in case the invariants
        # don't match.
        if (atm_obj.get_atomic_invariants() != invariants
                or is_aromatic is not None):
            self._events["modatms"].add(((atm_obj, pdb_atm),
                                         tuple(invariants),
                                         is_aromatic))

    def _fix_bonds(self, bond_types):
        # bond_types becomes an empty list if bond_types is None.
        bond_types = bond_types or []

        # If bond_types is an empty list, do nothing.
        if len(bond_types) == 0:
            return

        # It expects a list of tuples where the first element is the bond to be
        # updated, the second element should be the new type, and the third
        # optional element is a flag to indicate if the bond is aromatic
        # or not.
        for bond_info in bond_types:
            bond_obj, new_bond_type = bond_info[0:2]
            # Update bond type
            bond_obj.set_bond_type(new_bond_type)

            # If the tuple contains three elements, it must be a boolean to
            # define if it is an aromatic bond or not.
            if len(bond_info) == 3:
                is_aromatic = bond_info[2]
                bond_obj.set_as_aromatic(is_aromatic)

    def _add_bonds(self, mol_obj, new_bonds):
        # new_bonds becomes an empty list if new_bonds is None.
        new_bonds = new_bonds or []

        # If new_bonds is an empty list, do nothing.
        if len(new_bonds) == 0:
            return

        obmol = mol_obj
        if isinstance(mol_obj, MolWrapper):
            mol_obj = mol_obj.unwrap()

        # It expects a list of tuples where the first element is the bond to be
        # updated, the second element should be the new type, and the third
        # optional element is a flag to indicate if the bond is aromatic
        # or not.
        for bond_info in new_bonds:
            begin_idx, end_idx = bond_info[0:2]
            bond_type = bond_info[2]

            # If 'flags' is:
            #   - 0: the bond is not aromatic.
            #   - 2: the bond is aromatic.
            flags = 0

            # If the tuple contains four elements, it must be a boolean to
            # define if it is an aromatic bond or not.
            if len(bond_info) == 4:
                flags = 2 if bond_info[3] is True else 0

            obmol.AddBond(begin_idx, end_idx,
                          bond_type, flags)

    def _remove_explicit_hydrogens(self, atm_obj):
        delete_hs = []
        for b in atm_obj.get_bonds():
            if b.get_partner_atom(atm_obj).get_symbol() == "H":
                delete_hs.append(b.get_partner_atom(atm_obj))

        for hs_obj in delete_hs:
            atm_obj.parent.unwrap().DeleteAtom(hs_obj.unwrap())

    def _fix_atom(self, atm_obj, charge=None, remove_explict_h=True,
                  implicit_h_count=None, is_aromatic=None,
                  in_ring=None):

        # Remove all current explicit hydrogens.
        if remove_explict_h:
            self._remove_explicit_hydrogens(atm_obj)

        # All aromatic atoms will have their in_ring property set to True.
        if is_aromatic:
            in_ring = True

        # Set if atom belongs to a ring or not.
        if in_ring is not None:
            atm_obj.set_in_ring(in_ring)

        # Set if an atom is aromatic or not.
        if is_aromatic is not None:
            atm_obj.set_as_aromatic(is_aromatic)

        if implicit_h_count is not None:
            atm_obj.unwrap().SetImplicitHCount(implicit_h_count)

        # Set charge.
        if charge is not None:
            atm_obj.set_charge(charge)

    def _apply_modifications(self):

        processed_bonds = set()
        for data in self._events["break"]:
            atm_obj, pdb_atm = data[0]
            partner_obj, partner_atm = data[1]
            bond_to_break = data[2]

            bond_key = tuple(sorted([pdb_atm, partner_atm]))

            if bond_key in processed_bonds:
                continue
            if pdb_atm in self._events["ignore"]:
                continue

            if xor(pdb_atm.parent.is_metal(),
                   partner_atm.parent.is_metal()):

                if pdb_atm.parent.is_metal():
                    key, idx = partner_atm, partner_obj.get_idx()
                    val = pdb_atm.parent
                else:
                    key, idx = pdb_atm, atm_obj.get_idx()
                    val = partner_atm.parent

                coords_dict = self.metals_coord[key.parent][key]
                coords_dict["atm_idx"] = idx
                coords_dict["metals"].add(val)

            obmol = atm_obj.parent.unwrap()
            obmol.DeleteBond(bond_to_break.unwrap())

            processed_bonds.add(bond_key)

        for data in self._events["modbonds"]:
            (atm_obj, pdb_atm) = data[0]
            (partner_obj, partner_atm) = data[1]

            bond_to_mod, new_bond_type, is_aromatic = data[2:]

            if (pdb_atm in self._events["ignore"]
                    or partner_atm in self._events["ignore"]):
                continue

            if bond_to_mod is not None:
                self._fix_bonds([(bond_to_mod, new_bond_type, is_aromatic)])
            else:
                obmol = atm_obj.parent
                new_bond = [(atm_obj.get_idx(), partner_obj.get_idx(),
                             new_bond_type.value, is_aromatic)]
                self._add_bonds(obmol, new_bond)

        for data in self._events["modatms"]:
            (atm_obj, pdb_atm) = data[0]
            invariants = data[1]
            is_aromatic = data[2]

            if pdb_atm in self._events["ignore"]:
                continue

            charge, implicit_h_count, in_ring = None, None, None
            if invariants is not None:
                charge = invariants[4]
                implicit_h_count = invariants[5]
                in_ring = bool(invariants[6])

            self._fix_atom(atm_obj,
                           charge=charge,
                           implicit_h_count=implicit_h_count,
                           is_aromatic=is_aromatic,
                           in_ring=in_ring)

    def _get_precomp_data_from_cif_file(self, res):
        # TODO
        return {}

    def _check_mol(self, res):
        atom_pairs = self._comps[res]
        atm_names = [pdb_atm.name
                     for atm_obj, pdb_atm in atom_pairs]

        precomp_data = None
        if res.is_hetatm() or res.is_nucleotide():
            precomp_data = self._get_precomp_data_from_cif_file(res)

        self._validate_atoms_list(res, atm_names, precomp_data)

        resname = res.resname
        if res.is_residue():
            resname = self._resolve_resname(res)
        elif res.is_water():
            resname = "HOH"

        for atm_obj, pdb_atm in atom_pairs:
            if res.is_metal():
                self._check_metal_coordination(atm_obj, pdb_atm)

            if not res.is_hetatm() and not res.is_nucleotide():
                self._resolve_bonds(resname, atm_obj, pdb_atm)

            self._resolve_invariants(resname, atm_obj, pdb_atm,
                                     atm_names, precomp_data)

    def standardize(self, atom_pairs, metals_coord=None):

        if not metals_coord:
            custom_dict = lambda: {"atm_idx": None, "metals": set()}
            metals_coord = defaultdict(lambda: defaultdict(custom_dict))
        self.metals_coord = metals_coord

        self._comps = defaultdict(list)
        self._events = defaultdict(set)
        self._pdb_mapping = {}

        for atm_obj, pdb_atm in atom_pairs:
            self._pdb_mapping[atm_obj.get_idx()] = pdb_atm
            self._comps[pdb_atm.parent].append((atm_obj, pdb_atm))

            coords = self.metals_coord.get(pdb_atm.parent, {})
            if pdb_atm in coords:
                coords[pdb_atm]["atm_idx"] = atm_obj.get_idx()

        for res in self._comps:
            self._check_mol(res)

        self._apply_modifications()



import re
import logging
from operator import xor
from os.path import exists
from collections import defaultdict
import ast

from rdkit.Chem import Mol as RDMol
from openbabel import OBMol
from openbabel.pybel import readfile
from openbabel.pybel import Molecule as PybelMol
from openbabel.pybel import informats as OB_FORMATS

from luna.wrappers.rdkit import (RDKIT_FORMATS, read_multimol_file,
                                 read_mol_from_file)
from luna.wrappers.base import MolWrapper
from luna.util.default_values import (ACCEPTED_MOL_OBJ_TYPES, ENTRY_SEPARATOR,
                                      ARTIFACTS_LIST)
from luna.util.file import get_file_format, get_filename
from luna.util.exceptions import (InvalidEntry,
                                  IllegalArgumentError,
                                  MoleculeObjectError,
                                  MoleculeObjectTypeError,
                                  MoleculeNotFoundError)
from luna.pdb.parser.base import PDBParser, WATER_NAMES, DEFAULT_CHAIN_ID
from luna.pdb.core.entity import Entity


logger = logging.getLogger()

PCI_ENTRY_REGEX = re.compile(r'^.{1,255}:\w:\w[\w+\-]{0,2}:'
                             r'\-?\d{1,4}[a-zA-z]?$')
PPI_ENTRY_REGEX = re.compile(r'^.{1,255}:\w$')

REGEX_RESNUM_ICODE = re.compile(r'^([\-\+]?\d+)([a-zA-z]?)$')


class Entry:
    """Entries determine the reference molecule to which interactions and other
    properties will be calculated. They can be ligands, chains, etc, and can
    be defined in a number of ways. Each entry has an associated PDB file that
    may contain macromolecules (protein, RNA, DNA) and other small molecules,
    water, and ions. The PDB file provides the context to where the
    interactions with the reference molecule will be calculated.

    Parameters
    ----------
    pdb_id : str
        A 4-symbols structure id from PDB or a local PDB filename.
        Example: '3QL8' or 'file1'.
    chain_id : str
        A 1-symbol chain id. Example: 'A'.
    comp_name : str, optional
        A 1 to 3-symbols molecule name (residue name in the PDB format).
        Obligatory if ``is_hetatm`` is True. Example: 'X01'.
    comp_num : int, optional
        A valid 4-digits integer (residue sequence number in the PDB format).
        Obligatory if ``is_hetatm`` is True. Example: 300 or -1.
    comp_icode : str, optional
        A 1-character molecule insertion code (residue insertion code in the
        PDB format). Example: 'A'.
    is_hetatm : bool
        If the molecule is a ligand or not. The default value is True.
    sep : str
        A separator character to format the entry string.
        The default value is ':'.
    parser : :class:`~luna.pdb.parser.base.PDBParser` or \
                :class:`~luna.pdb.parser.ftmap.FTMapParser`, optional
            Define a PDB parser object. If not provided, the default parser
            will be used.

    Raises
    ------
    IllegalArgumentError
        If ``is_hetatm`` is True, but the molecule name and number are
        not provided. If the molecule number is provided but it is not
        an integer. If ``comp_icode`` is provided but it is not a
        valid character.
    InvalidEntry
        If the provided information does not match the PDB format.

    Examples
    --------

    Chain entry: can be used to calculate interactions with a given chain.

    >>> from luna.mol.entry import Entry
    >>> e = Entry(pdb_id="3QL8", chain_id="A")
    >>> print(e)
    <Entry: 3QL8:A>

    Molecule entry: can be used to calculate interactions with a given molecule
    (residue or nucleotide).

    >>> from luna.mol.entry import Entry
    >>> e = Entry(pdb_id="3QL8", chain_id="A", comp_name="HIS", comp_num=125, \
is_hetatm=False)
    >>> print(e)
    <Entry: 3QL8:A:HIS:125>

    Ligand entry: can be used to calculate interactions with a given ligand.

    >>> from luna.mol.entry import Entry
    >>> e = Entry(pdb_id="3QL8", chain_id="A", comp_name="X01", comp_num=300, \
is_hetatm=True)
    >>> print(e)
    <Entry: 3QL8:A:X01:300>

    You can use a different character separator for the entries. For example:

    >>> from luna.mol.entry import Entry
    >>> e = Entry(pdb_id="3QL8", chain_id="A", comp_name="X01", comp_num=300, \
is_hetatm=True, sep="/")
    >>> print(e)
    <Entry: 3QL8/A/X01/300>
    """

    def __init__(self, pdb_id, chain_id, comp_name=None,
                 comp_num=None, comp_icode=None,
                 is_hetatm=True, sep=ENTRY_SEPARATOR, parser=None):

        if xor(comp_name is None, comp_num is None):
            raise IllegalArgumentError("You tried to define a molecule, so "
                                       "you must inform its name and number.")

        if comp_num is not None:
            try:
                assert float(comp_num).is_integer()
                comp_num = int(comp_num)
            except (ValueError, AssertionError):
                error_msg = ("The informed molecule number '%s' is invalid. "
                             "It must be an integer." % str(comp_num))
                raise IllegalArgumentError(error_msg)

        if comp_icode is not None:
            comp_icode = str(comp_icode)
            if comp_icode.isdigit() or len(comp_icode) > 1:
                error_msg = ("The informed molecule icode '%s' is invalid. "
                             "It must be a character." % str(comp_icode))
                raise IllegalArgumentError(error_msg)

        self._pdb_id = pdb_id
        self._chain_id = chain_id
        self._comp_name = comp_name
        self._comp_num = comp_num
        self._comp_icode = comp_icode
        self.is_hetatm = is_hetatm
        self.sep = sep
        self.parser = parser

        if not self.is_valid():
            raise InvalidEntry("Entry '%s' does not match the PDB format."
                               % self.to_string())

    @classmethod
    def from_string(cls, entry_str, is_hetatm=True, sep=ENTRY_SEPARATOR):
        """Initialize from a string.

        Parameters
        ----------
        entry_str : str
            A string representing the entry. Example: '3QL8:A:X01:300'.
        is_hetatm : bool
            Defines if the molecule is a ligand or not.
            The default value is True.
        sep : str
            The separator character used in ``entry_str``.
            The default value is ':'. For example: if ``entry_str`` is set to
            '3QL8|A|X01|300', then ``sep`` should be defined as '|'.

        Returns
        -------
         : `Entry`

        Raises
        ------
        IllegalArgumentError
            If the fields in ``entry_str`` do not match the format expected to
            define a chain (`ChainEntry`) or a molecule (`MolEntry`).

        Examples
        --------

        Chain entry: can be used to calculate interactions with a given chain.

        >>> from luna.mol.entry import Entry
        >>> e = Entry.from_string("3QL8:A", sep=":")
        >>> print(e)
        <Entry: 3QL8:A>

        Molecule entry: can be used to calculate interactions with a given
        molecule (residue or nucleotide).

        >>> from luna.mol.entry import Entry
        >>> e = Entry.from_string("3QL8:A:HIS:125", sep=":")
        >>> print(e)
        <Entry: 3QL8:A:HIS:125>

        Ligand entry: can be used to calculate interactions with a given
        ligand.

        >>> from luna.mol.entry import Entry
        >>> e = Entry.from_string("3QL8:A:X01:300", sep=":")
        >>> print(e)
        <Entry: 3QL8:A:X01:300>
        """

        entries = entry_str.split(sep)

        # Try to initialize a new ChainEntry.
        if len(entries) == 2:
            if any([str(i).strip() == "" for i in entries]):
                error_msg = ("The number of fields in the informed string "
                             "'%s' is incorrect. A valid ChainEntry must "
                             "contain two obligatory fields: PDB and chain "
                             "id." % entry_str)
                raise IllegalArgumentError(error_msg)

            return cls(*entries, is_hetatm=False, sep=sep)

        # Try to initialize a new MolEntry.
        elif len(entries) == 4:
            if any([str(i).strip() == "" for i in entries]):
                error_msg = ("The number of fields in the informed string "
                             "'%s' is incorrect. A valid MolEntry must "
                             "contain four obligatory fields: PDB, chain id, "
                             "molecule name, and molecule number followed by "
                             "its insertion code when applicable." % entry_str)
                raise IllegalArgumentError(error_msg)

            # Separate ligand number from insertion code.
            matched = REGEX_RESNUM_ICODE.match(entries[3])
            if matched:
                comp_num = matched.group(1)
                try:
                    assert float(comp_num).is_integer()
                    comp_num = int(comp_num)
                except (ValueError, AssertionError):
                    error_msg = ("The informed molecule number '%s' is "
                                 "invalid. It must be an integer."
                                 % str(comp_num))
                    raise IllegalArgumentError(error_msg)

                icode = None if matched.group(2) == "" else matched.group(2)
                entries = entries[0:3] + [comp_num, icode]
            else:
                error_msg = ("The molecule number and its insertion code "
                             "(if applicable) '%s' is invalid. It must be an "
                             "integer followed by one insertion code "
                             "character when applicable." % entries[3])
                raise IllegalArgumentError(error_msg)

            return cls(*entries, is_hetatm=is_hetatm, sep=sep)
        else:
            error_msg = ("The number of fields in the informed string '%s' is "
                         "incorrect. A valid string must contain two "
                         "obligatory fields (PDB and chain id) and it may "
                         "contain two optional fields (molecule name and "
                         "molecule number followed by its insertion code when "
                         "applicable)." % entry_str)
            raise IllegalArgumentError(error_msg)

    @classmethod
    def from_file(cls, input_file, pdb_id=None, mol_file=None,
                  entries_sep=":", fields_sep=",", **kwargs):
        """Initialize from a file containing a list of strings
        representing entries.

        Parameters
        ----------
        input_file : str
            The file from where the list of strings (one per line) will
            be read from.

            .. note::
                The list of strings can be a mix of complexes whose
                molecules lie all in the same PDB file, or whose protein
                structure is in a PDB file and whose ligand is on a
                molecular file. See ``entries_sep`` and ``fields_sep``.
        pdb_id : str
            A 4-symbols structure id from PDB or a local PDB filename.
            Example: '3QL8' or 'file1'.

            .. note::
                Mandatory if there is any line containing only the ligand name.
                This only applies to ligands that will be read from
                multimolecular files.
        mol_file : str
            Pathname of the molecular file.

            .. note::
                Mandatory if there is any line containing only the ligand name.
                This only applies to ligands that will be read from
                multimolecular files.
        entries_sep : str
            a separator for entries. For complexes where the ligand and the 
            protein structure lie all in the same PDB file (e.g., complexes 
            from the RCSB PDB), the separator must be the same used for 
            entries provided at `input_file`. 
            Default: ':'. E.g.: 3QQK:A:X02:497 or receptor:ZINC000649623159

        fields_sep : str
            a character used to separate fields in ``input_file`` for
            complexes whose protein structure is in a PDB file and
            whose ligand is on a molecular file. Such entries cause the
            function to yield `MolFileEntry` objects. Default: ','

            Each line should contain either the ligand name only or the
            following fields:

                * ``pdb_id`` (str): a 4-symbols structure id from \
PDB or a local PDB filename. Example: '3QL8' or 'file1'.
                * ``mol_id`` (str): the ligand id in the molecular file.

                * ``mol_file`` (str): pathname of the molecular file.

                * ``is_multimol_file`` (bool): if ``mol_file`` contains \
multiple molecules or not. If True, ``mol_id`` should match some ligand name \
in ``mol_file``.

            .. note::
                If a line contains only the ligand name, providing a value to
                the arguments ``pdb_id`` and ``mol_file`` is mandatory.
        kwargs : dict, optional
            Extra arguments to lines that yield `MolFileEntry` objects.
            Refer to the documentation for a list of all possible arguments.

        Yields
        ------
        `ChainEntry`, `MolEntry`, `MolFileEntry`
            Define a chain, a molecule from a PDB file, or a molecule from
            a molecular file, respectively, according to a string
            from ``input_file``.

        Raises
        ------
        IllegalArgumentError
            If a given line contain a string that do not match the format
            expected to define a chain (`ChainEntry`) or a molecule from a PDB
            file (`MolEntry`), or a molecule from a molecular file
            (`MolFileEntry`).

        Examples
        --------

        Example of a file containing a mix of entry types:

        >>> 3QQK:A:X02:497
        >>> ZINC000065293174
        >>> protein,ZINC000012442563,inputs/ligands.mol2,True
        >>> protein,ZINC000007786517,inputs/ZINC000007786517.mol,False
        """
        if not exists(input_file):
            raise OSError("File '%s' does not exist." % input_file)

        with open(input_file, "r") as IN:
            c = 1
            for row in IN:
                row = row.strip()
                if row == "":
                    continue

                args = row.split(fields_sep)

                has_error = False
                if len(args) == 1:
                    args = row.split(entries_sep)

                    if len(args) == 1:
                        if pdb_id is None or mol_file is None:
                            msg = ("It seems a molecule name was provided "
                                   f"in line #{c}. In this case, "
                                   "'pdb_id' and 'mol_file' are mandatory.")
                            raise IllegalArgumentError(msg)

                        yield MolFileEntry.from_mol_file(pdb_id, args[0],
                                                         mol_file,
                                                         is_multimol_file=True,
                                                         sep=entries_sep,
                                                         **kwargs)

                    elif len(args) == 2:
                        yield ChainEntry.from_string(row, sep=entries_sep)

                    elif len(args) == 4:
                        yield MolEntry.from_string(row, sep=entries_sep)

                    else:
                        has_error = True

                elif len(args) == 4:
                    (curr_pdb_id, curr_mol_id,
                        curr_mol_file, is_multimol) = args
                    is_multimol = ast.literal_eval(is_multimol)

                    cls = MolFileEntry
                    yield cls.from_mol_file(curr_pdb_id, curr_mol_id,
                                            curr_mol_file,
                                            is_multimol_file=is_multimol,
                                            sep=entries_sep,
                                            **kwargs)
                else:
                    has_error = True

                if has_error:
                    msg = (f"Invalid number of arguments found in line #{c}. "
                           "You should provide either one or four arguments. "
                           "If only one argument is provided, it can be an "
                           "entry string (e.g., 3QQK:A:X02:497) or a "
                           "molecule name.")
                    raise IllegalArgumentError(msg)

                c += 1

    @property
    def pdb_id(self):
        """str, read-only: the pdb id."""
        return self._pdb_id

    @property
    def chain_id(self):
        """str, read-only: the chain id."""
        return self._chain_id

    @property
    def comp_name(self):
        """str, read-only: the molecule name."""
        return self._comp_name

    @property
    def comp_num(self):
        """int, read-only: the molecule number."""
        return self._comp_num

    @property
    def comp_icode(self):
        """str, read-only: the molecule insertion code."""
        if isinstance(self._comp_icode, str):
            return self._comp_icode
        else:
            return ' '

    @property
    def full_id(self):
        """tuple, read-only: The full id of the entry is the tuple
        (PDB id or filename, chain id) for entries representing chains and
        (PDB id or filename, chain id, molecule name, molecule number,
        insertion code) for entries representing molecules."""

        entry = [self.pdb_id, self.chain_id]
        if self.comp_name is not None and self.comp_num is not None:
            entry.append(self.comp_name)
            entry.append(self.comp_num)
            entry.append(self.comp_icode)
        return tuple(entry)

    def to_string(self, sep=None):
        """ Convert the entry to a string using ``sep`` as a
        separator character.

        Parameters
        ----------
        sep : str or None
            If None (the default), use the separator character defined during
            the entry object creation. Otherwise, uses ``sep`` as the
            separator character.

        Examples
        --------

        >>> from luna.mol.entry import Entry
        >>> e = Entry(pdb_id="3QL8", chain_id="A", comp_name="X01", \
comp_num=300, is_hetatm=True, sep=":")
        >>> print(e.to_string("/"))
        3QL8/A/X01/300
        """

        full_id = self.full_id

        # An entry object will always have a PDB and chain id.
        entry = list(full_id[0:2])

        # If it contains additional information about the molecule it
        # will also include them.
        if len(full_id) > 2:
            if full_id[2] is not None and full_id[3] is not None:
                comp_name = str(full_id[2]).strip()
                comp_num_and_icode = \
                    str(full_id[3]).strip() + str(full_id[4]).strip()
                entry += [comp_name, comp_num_and_icode]

        sep = sep or self.sep

        return sep.join(entry)

    def is_valid(self):
        """Check if the entry matches the expected format for protein-protein
        or protein-molecule complexes.

        Returns
        -------
         : bool
        """
        full_id = self.full_id

        # Regex for ChainEntry (pdb_id, chain_id).
        if len(full_id) == 2:
            return PPI_ENTRY_REGEX.match(self.to_string(":")) is not None

        # Regex for MolEntry (pdb_id, chain_id, comp_name, comp_num, icode).
        elif len(full_id) == 5:
            return PCI_ENTRY_REGEX.match(self.to_string(":")) is not None

        # Return False for anything else
        return False

    def get_biopython_key(self, full_id=False):
        """Represent the entry as a key to select chains or molecules from
        Biopython Entity objects.

        Parameters
        full_id : bool
            If True, return the full id of a chain or ligand.
            For chains, it consists of a tuple containing the PDB and the
            chain id. For ligands, it consists of a tuple containing the PDB,
            the chain, and the ligand id. The default value is False.

        Returns
        -------
         : str or tuple
            Return str if the entry represents a chain and if ``full_id``
            is False. Otherwise, return a tuple.

        Examples
        --------

        >>> from luna.mol.entry import Entry
        >>> e = Entry(pdb_id="3QL8", chain_id="A", comp_name="X01", \
comp_num=300, is_hetatm=True, sep=":")
        >>> print(e.get_biopython_key())
        ('H_X01', 300, ' ')

        """
        key = []
        if full_id:
            key = [self.pdb_id, 0, self.chain_id]

        if self.comp_name is not None and self.comp_num is not None:
            if self.comp_name == 'HOH' or self.comp_name == 'WAT':
                comp_id = ('W', self.comp_num, self.comp_icode)
            elif self.is_hetatm:
                comp_id = ('H_%s' % self.comp_name, self.comp_num,
                           self.comp_icode)
            else:
                comp_id = (' ', self.comp_num, self.comp_icode)

            if full_id:
                key.append(comp_id)
                return tuple(key)
            return comp_id

        if full_id:
            return tuple(key)

        return self.chain_id

    def __repr__(self):
        return '<%s: %s>' % (self.__class__.__name__, self.to_string(self.sep))


class ChainEntry(Entry):
    """Define a chain.

    Parameters
    ----------
    pdb_id : str
        A 4-symbols structure id from PDB or a local PDB filename.
        Example: '3QL8' or 'file1'.
    chain_id : str
        A 1-symbol chain id. Example: 'A'.
    sep : str
        A separator character to format the entry string.
        The default value is ':'.

    Raises
    ------
    InvalidEntry
        If the provided information does not match the PDB format.

    Examples
    --------
    >>> from luna.mol.entry import ChainEntry
    >>> e = ChainEntry(pdb_id="3QL8", chain_id="A")
    >>> print(e)
    <ChainEntry: 3QL8:A>

    """

    def __init__(self, pdb_id, chain_id, sep=ENTRY_SEPARATOR, parser=None):
        super().__init__(pdb_id, chain_id, is_hetatm=False, sep=sep,
                         parser=parser)

    @classmethod
    def from_string(cls, entry_str, sep=ENTRY_SEPARATOR):
        """Initialize from a string.

        Parameters
        ----------
        entry_str : str
            A string representing the entry. Example: '3QL8:A'.
        sep : str
            The separator character used in ``entry_str``.
            The default value is ':'. For example: if ``entry_str`` is set to
            '3QL8|A', then ``sep`` should be defined as '|'.

        Returns
        -------
         : `Entry`

        Raises
        ------
        IllegalArgumentError
            If the fields in ``entry_str`` do not match the format expected
            to define a chain.

        Examples
        --------
        >>> from luna.mol.entry import ChainEntry
        >>> e = ChainEntry.from_string("3QL8:A", sep=":")
        >>> print(e)
        <ChainEntry: 3QL8:A>

        """
        entries = entry_str.split(sep)
        if len(entries) == 2:
            return cls(*entries, sep=sep)
        else:
            error_msg = ("The number of fields in the informed string '%s' is "
                         "incorrect. A valid string must contain two "
                         "obligatory fields: PDB and chain id." % entry_str)
            raise IllegalArgumentError(error_msg)

    @property
    def full_id(self):
        """tuple, read-only: The full id of the entry is the tuple
        (PDB id or filename, chain id)."""
        return (self.pdb_id, self.chain_id)


class MolEntry(Entry):
    """Define a molecule from a PDB file, which can be a residue,
    nucleotide, ligand, water, or ion.

    Parameters
    ----------
    pdb_id : str
        A 4-symbols structure id from PDB or a local PDB filename.
        Example: '3QL8' or 'file1'.
    chain_id : str
        A 1-symbol chain id. Example: 'A'.
    comp_name : str
        A 1 to 3-symbols molecule name (residue name in the PDB format).
        Example: 'X01'.
    comp_num : int
        A valid 4-digits integer (residue sequence number in the PDB format).
        Example: 300 or -1.
    comp_icode : str, optional
        A 1-character molecule insertion code (residue insertion code in the
        PDB format). Example: 'A'.
    sep : str
        A separator character to format the entry string.
        The default value is ':'.

    Raises
    ------
    InvalidEntry
        If the provided information does not match the PDB format.

    Examples
    --------

    Molecule entry: can be used to calculate interactions with a given
    molecule (residue or nucleotide).

    >>> from luna.mol.entry import MolEntry
    >>> e = MolEntry(pdb_id="3QL8", chain_id="A", comp_name="HIS", \
comp_num=125, is_hetatm=False)
    >>> print(e)
    <MolEntry: 3QL8:A:HIS:125>

    Ligand entry: can be used to calculate interactions with a given ligand.

    >>> from luna.mol.entry import MolEntry
    >>> e = MolEntry(pdb_id="3QL8", chain_id="A", comp_name="X01", \
comp_num=300, is_hetatm=True)
    >>> print(e)
    <MolEntry: 3QL8:A:X01:300>

    """

    def __init__(self, pdb_id, chain_id, comp_name, comp_num,
                 comp_icode=None, sep=ENTRY_SEPARATOR, parser=None):

        super().__init__(pdb_id, chain_id, comp_name, comp_num, comp_icode,
                         is_hetatm=True, sep=sep, parser=parser)

    @classmethod
    def from_string(cls, entry_str, sep=ENTRY_SEPARATOR):
        entries = entry_str.split(sep)
        if len(entries) == 4:
            # Separate ligand number from insertion code.
            matched = REGEX_RESNUM_ICODE.match(entries[3])
            if matched:
                comp_num = matched.group(1)
                icode = None if matched.group(2) == "" else matched.group(2)
                entries = entries[0:3] + [comp_num, icode]
            else:
                error_msg = ("The molecule number and its insertion code "
                             "(if applicable) '%s' is invalid. It must be an "
                             "integer followed by one insertion code "
                             "character when applicable." % entries[3])
                raise IllegalArgumentError(error_msg)
            return cls(*entries, sep=sep)
        else:
            error_msg = ("The number of fields in the informed string '%s' is "
                         "incorrect. A valid molecule entry must contain four "
                         "obligatory fields: PDB, chain id, molecule name, "
                         "and molecule number followed by its insertion code "
                         "when applicable." % entry_str)
            raise IllegalArgumentError(error_msg)


class MolFileEntry(Entry):

    """Define a ligand from a molecular file.
    This class should be used for docking and molecular dynamics campaigns
    where usually one has the protein structure in the PDB format and the
    ligand structure in a separate molecular file.

    Parameters
    ----------
    pdb_id : str
        A 4-symbols structure id from PDB or a local PDB filename.
        Example: '3QL8' or 'file1'.
    mol_id : str
        The ligand id in the molecular file.
    sep : str
        A separator character to format the entry string.
        The default value is ':'.

    Attributes
    ----------
    mol_id : str
        The ligand id.
    mol_file : str
        Pathname of the molecular file.
    mol_file_ext : str
        The molecular file format.
        If not provided, try to recover the molecular file extension
        directly from ``mol_file``.
    mol_obj_type : {'rdkit', 'openbabel'}
        Define which library (RDKit or Open Babel) to use to parse
        the molecular file.
    overwrite_mol_name : bool
        If True, substitute the ligand name in the parsed molecular object
        with ``mol_id``. Only works for single-molecule files
        (``is_multimol_file`` = False) as in these cases ``mol_id`` does not
        need to match the ligand name in the molecular file.
    is_multimol_file : bool
        If ``mol_file`` contains multiple molecules or not.
        If True, ``mol_id`` should match some ligand name in ``mol_file``.
    """

    def __init__(self, pdb_id, mol_id, sep=ENTRY_SEPARATOR):

        self.mol_id = mol_id

        #
        # Initialize empty properties.
        #
        self._mol_obj = None
        self.mol_file = None
        # TODO: Find a way to assume the Mol file type when not provided
        self.mol_file_ext = None
        self.mol_obj_type = None
        self.overwrite_mol_name = None
        self.is_multimol_file = None

        super().__init__(pdb_id, DEFAULT_CHAIN_ID, "LIG", 9999,
                         is_hetatm=True, sep=sep)

    @classmethod
    def from_mol_obj(cls, pdb_id, mol_id, mol_obj, sep=ENTRY_SEPARATOR):
        """Initialize from an already loaded molecular object.

        This function is useful in cases where a molecular object is parsed and
        pre-processed using a different protocol defined by the user.

        Parameters
        ----------
        pdb_id : str
            A 4-symbols structure id from PDB or a local PDB filename.
            Example: '3QL8' or 'file1'.
        mol_id : str
            The ligand id.
            As the molecular object is already provided, the ligand id does
            not need to match the ligand name in the molecular object.
        mol_obj : :class:`~luna.wrappers.base.MolWrapper`, \
                    :class:`rdkit.Chem.rdchem.Mol`, or \
                    :class:`openbabel.pybel.Molecule`
            The molecular object.
        sep : str
            A separator character to format the entry string.
            The default value is ':'.

        Returns
        -------
         : `MolFileEntry`

        Raises
        ------
        MoleculeObjectTypeError
            If the molecular object is not an instance
            of :class:`~luna.wrappers.base.MolWrapper`,
            :class:`rdkit.Chem.rdchem.Mol`,
            or :class:`openbabel.pybel.Molecule`.
        IllegalArgumentError
            If ``entity`` is not a valid Biopython object.

        Examples
        --------

        In this example, we will initialize a new MolFileEntry with the ligand
        'ZINC000007786517' and the structure located in a PDB file of name
        'D4', which is the structure used for docking the molecule.

        First, let's parse the molecular file.

        >>> from luna.wrappers.rdkit import read_mol_from_file
        >>> mol_file = "tutorial/inputs/ZINC000007786517.mol"
        >>> mol_obj = read_mol_from_file(mol_file, mol_format="mol")

        Now, we create the new MolFileEntry object as follows:

        >>> e = MolFileEntry.from_mol_obj("D4", "ZINC000007786517",
        ...                               mol_obj, sep=ENTRY_SEPARATOR)
        >>> print(e)
        <MolFileEntry: D4:ZINC000007786517>
        >>> print(e.mol_obj.to_smiles())
        Cc1cccc(NC(=O)C[N@@H+](C)C2CCCCC2)c1C
        """

        if mol_obj is not None:
            if isinstance(mol_obj, MolWrapper):
                mol_obj = mol_obj.unwrap()
            elif isinstance(mol_obj, PybelMol):
                mol_obj = mol_obj.OBMol

            if isinstance(mol_obj, RDMol):
                mol_obj_type = "rdkit"
            elif isinstance(mol_obj, OBMol):
                mol_obj_type = "openbabel"
            else:
                logger.exception("Objects of type '%s' are not currently "
                                 "accepted." % mol_obj.__class__)
                raise MoleculeObjectTypeError("Objects of type '%s' are not "
                                              "currently accepted."
                                              % mol_obj.__class__)
        else:
            if mol_obj_type not in ACCEPTED_MOL_OBJ_TYPES:
                error_msg = ("Objects of type '%s' are not currently "
                             "accepted. The available options are: %s."
                             % (mol_obj_type,
                                ", ".join(ACCEPTED_MOL_OBJ_TYPES)))
                raise IllegalArgumentError(error_msg)

        entry = cls(pdb_id, mol_id, sep)
        entry.mol_obj = mol_obj
        entry.mol_obj_type = mol_obj_type

        # TODO: Find a way to assume the Mol file type when not provided
        # mol_file_ext

        return entry

    @classmethod
    def from_mol_file(cls, pdb_id, mol_id, mol_file, is_multimol_file,
                      mol_file_ext=None, mol_obj_type='rdkit',
                      autoload=False, overwrite_mol_name=False,
                      sep=ENTRY_SEPARATOR):
        """Initialize from a molecular file.

        Parameters
        ----------
        pdb_id : str
            A 4-symbols structure id from PDB or a local PDB filename.
            Example: '3QL8' or 'file1'.
        mol_id : str
            The ligand id in the molecular file.
        mol_file : str
            Pathname of the molecular file.
        is_multimol_file : bool
            If ``mol_file`` contains multiple molecules or not.
            If True, ``mol_id`` should match some ligand name in ``mol_file``.
        mol_file_ext : str, optional
            The molecular file format.
            If not provided, try to recover the molecular file extension
            directly from ``mol_file``.
        mol_obj_type : {'rdkit', 'openbabel'}
            If "rdkit", parse the converted molecule with RDKit and return an
            instance of :class:`rdkit.Chem.rdchem.Mol`. If "openbabel", parse
            the converted molecule with Open Babel and return an instance of
            :class:`openbabel.pybel.Molecule`. The default value is 'rdkit'.
        autoload : bool
            If True, parse the ligand from the molecular file during the entry
            initialization. Otherwise, only load the ligand when first used.
        overwrite_mol_name : bool
            If True, substitute the ligand name in the parsed molecular object
            with ``mol_id``. Only works for single-molecule files
            (``is_multimol_file`` = False) as in these cases ``mol_id`` does
            not need to match the ligand name in the molecular file.
        sep : str
            A separator character to format the entry string.
            The default value is ':'.

        Returns
        -------
         : `MolFileEntry`

        Raises
        ------
        FileNotFoundError
            If ``mol_file`` does not exist.
        IllegalArgumentError
            If ``mol_obj_type`` is not either 'rdkit' nor 'openbabel'.
        MoleculeObjectError
            If any errors occur while parsing the molecular file. Detailed
            information about the errors can be found in the logging outputs.
        MoleculeNotFoundError
            If the ligand ``mol_id`` was not found in the input file and
            ``is_multimol_file`` is True.

        Examples
        --------

        In this first example, we will read the ligand 'ZINC000007786517' from
        a single-molecule file. As we are working with a single-molecule file,
        ``mol_id`` can be any value you prefer.

        >>> from luna.mol.entry import MolFileEntry
        >>> mol_file = "tutorial/inputs/ZINC000007786517.mol"
        >>> e = MolFileEntry.from_mol_file(pdb_id="D4", mol_id="Ligand",
        ...                                mol_file=mol_file, \
mol_obj_type='rdkit',
        ...                                is_multimol_file=False)
        >>> print(e)
        D4:Ligand
        >>> print(e.mol_obj.to_smiles())
        Cc1cccc(NC(=O)C[N@@H+](C)C2CCCCC2)c1C

        Now, let's say we need to read the ligand 'ZINC000096459890' from a
        multi-molecular file and that we want to use Open Babel to parse the
        molecule. To do so, remember that it should exist a ligand with the
        name ``mol_id`` in ``mol_file``. Otherwise, it will raise the
        exception MoleculeNotFoundError.

        >>> from luna.mol.entry import MolFileEntry
        >>> mol_file = "tutorial/inputs/ligands.mol2"
        >>> e = MolFileEntry.from_mol_file(pdb_id="D4", \
mol_id="ZINC000096459890",
        ...                                mol_file=mol_file, \
mol_obj_type='openbabel',
        ...                                is_multimol_file=True)
        >>> print(e)
        <MolFileEntry: D4:ZINC000096459890>
        >>> print(e.mol_obj.to_smiles())
        O=C(OCCCN1C=CC=CC1=O)c1ccc2ccc(Cl)cc2n1

        Below, we show what happens if ``mol_id`` does not exist in
        ``mol_file``. Observe we set ``autoload`` to True to parse the
        molecule right away.

        >>> from luna.mol.entry import MolFileEntry
        >>> mol_file = "tutorial/inputs/ligands.mol2"
        >>> e = MolFileEntry.from_mol_file(pdb_id="D4", mol_id="Ligand",
        ...                                mol_file=mol_file, \
mol_obj_type='openbabel',
        ...                                 is_multimol_file=True, \
autoload=True)
        luna.util.exceptions.MoleculeNotFoundError: "The ligand 'Ligand' was \
not found in the
        input file or generated errors while parsing it with Open Babel."
        """

        entry = cls(pdb_id, mol_id, sep)

        entry.mol_file = mol_file
        entry.is_multimol_file = is_multimol_file
        entry.mol_file_ext = mol_file_ext or get_file_format(mol_file)
        entry.mol_obj_type = mol_obj_type
        entry.overwrite_mol_name = overwrite_mol_name

        if autoload:
            entry._load_mol_from_file()

        return entry

    @classmethod
    def from_file(cls, input_file, pdb_id=None, mol_file=None,
                  fields_sep=",", **kwargs):

        if not exists(input_file):
            raise OSError("File '%s' does not exist." % input_file)

        with open(input_file, "r") as IN:
            c = 1
            for row in IN:
                row = row.strip()
                if row == "":
                    continue

                args = row.split(fields_sep)

                if len(args) == 1:
                    if pdb_id is None or mol_file is None:
                        msg = (f"Only one field was found in line #{c}. "
                               "In these cases, 'pdb_id' and 'mol_file' "
                               "are mandatory.")
                        raise IllegalArgumentError(msg)

                    curr_pdb_id = pdb_id
                    curr_mol_id = args[0]
                    curr_mol_file = mol_file
                    is_multimol_file = True

                elif len(args) == 4:
                    (curr_pdb_id, curr_mol_id,
                        curr_mol_file, is_multimol_file) = args
                    is_multimol_file = ast.literal_eval(is_multimol_file)

                else:
                    msg = (f"Invalid number of arguments found in line #{c}. "
                           "You should provide either one or four arguments.")
                    raise IllegalArgumentError(msg)

                yield cls.from_mol_file(curr_pdb_id, curr_mol_id,
                                        curr_mol_file,
                                        is_multimol_file=is_multimol_file,
                                        **kwargs)
                c += 1

    @property
    def full_id(self):
        """tuple, read-only: The full id of the entry is the tuple
        (PDB id or filename, ligand id)."""
        return (self.pdb_id, self.mol_id)

    @property
    def mol_obj(self):
        """:class:`~luna.wrappers.base.MolWrapper`, \
        :class:`rdkit.Chem.rdchem.Mol`, \
        or :class:`openbabel.pybel.Molecule`: The molecule."""
        if self._mol_obj is None and self.mol_file is not None:
            self._load_mol_from_file()
        return self._mol_obj

    @mol_obj.setter
    def mol_obj(self, mol_obj):
        self._mol_obj = MolWrapper(mol_obj)

    def is_valid(self):
        """Check if the entry represents a valid protein-ligand complex.

        Returns
        -------
         : bool
        """
        return True

    def is_mol_obj_loaded(self):
        """Check if the molecular object has already been loaded.

        Returns
        -------
         : bool
        """
        return self._mol_obj is not None

    def _load_mol_from_file(self):
        logger.debug("It will try to load the molecule '%s'." % self.mol_id)

        if self.mol_file is None:
            raise IllegalArgumentError("It cannot load the molecule as no "
                                       "molecular file was provided.")

        available_formats = (OB_FORMATS if self.mol_obj_type == "openbabel"
                             else RDKIT_FORMATS)
        tool = "Open Babel" if self.mol_obj_type == "openbabel" else "RDKit"
        if self.mol_file_ext not in available_formats:
            raise IllegalArgumentError("Extension '%s' informed or assumed "
                                       "from the filename is not a format "
                                       "recognized by %s."
                                       % (self.mol_file_ext, tool))

        if not exists(self.mol_file):
            raise FileNotFoundError("The file '%s' was not found."
                                    % self.mol_file)

        try:
            if self.mol_obj_type == "openbabel":
                mols = readfile(self.mol_file_ext, self.mol_file)
                # If it is a multimol file, then we need to loop over the
                # molecules to find the target one. Note that in this case,
                # the ids must match.
                if self.is_multimol_file:
                    for ob_mol in mols:
                        filename = get_filename(ob_mol.OBMol.GetTitle())
                        if self.mol_id == filename:
                            self.mol_obj = ob_mol
                            break
                else:
                    self.mol_obj = mols.__next__()
            else:
                if self.mol_file_ext == "pdb":
                    self.mol_obj = \
                        read_mol_from_file(self.mol_file,
                                           mol_format=self.mol_file_ext,
                                           removeHs=False)
                else:
                    # If 'targets' is None, then the entire Mol file
                    # will be read.
                    targets = None
                    # If it is a multimol file than loop through it until the
                    # informed molecule (by its mol_id) is found.
                    if self.is_multimol_file:
                        targets = [self.mol_id]

                    for rdk_mol, mol_id \
                        in read_multimol_file(self.mol_file,
                                              mol_format=self.mol_file_ext,
                                              targets=targets,
                                              removeHs=False):
                        # It returns None if the molecule parsing
                        # generated errors.
                        self.mol_obj = rdk_mol
                        break
        except Exception as e:
            logger.exception(e)

            error_msg = ("An error occurred while parsing the molecular file "
                         "with %s and the molecule object for the entry '%s' "
                         "could not be created. Check the logs for more "
                         "information." % (tool, self.to_string()))
            raise MoleculeObjectError(error_msg)

        if self._mol_obj is None:
            error_msg = ("The ligand '%s' was not found in the input file or "
                         "generated errors while parsing it with %s."
                         % (self.mol_id, tool))
            raise MoleculeNotFoundError(error_msg)
        else:
            if not self.mol_obj.has_name() or self.overwrite_mol_name:
                self.mol_obj.set_name(self.mol_id)

        logger.debug("Molecule '%s' was successfully loaded." % self.mol_id)

    def get_biopython_structure(self, entity=None, parser=None):
        """Transform the molecular object into a Biopython Entity object.

        If ``entity`` is provided, the molecular object is appended to it,
        i.e., this function can be used to join a ligand and the structure
        used during docking or molecular dynamics.

        By default, the ligand is added to a chain of id `z`.

        Parameters
        ----------
        entity : :class:`~luna.pdb.core.entity.Entity`, optional
            Append the molecular object to ``entity``.
            If not provided, a new :class:`~luna.pdb.core.entity.Entity`
            is created.
        parser : :class:`~luna.pdb.parser.base.PDBParser`, optional
            Define a PDB parser object. If not provided, the default parser
            will be used.

        Returns
        -------
         : :class:`~luna.pdb.core.entity.Entity`

        Raises
        ------
        IllegalArgumentError
            If ``entity`` is not a valid Biopython object.

        Examples
        --------

        In this example, we will demonstrate how to join a protein structure
        and a ligand docked against it.

        First, let's parse the PDB file.

        >>> from luna.pdb.parser.base.PDBParser import PDBParser
        >>> pdb_parser = PDBParser(PERMISSIVE=True, QUIET=True)
        >>> pdb_file = "tutorial/inputs/D4.pdb"
        >>> structure = pdb_parser.get_structure("Protein", pdb_file)

        Observe that the list of chains in the parsed structure contains
        only one element.

        >>> print(structure[0].child_list)
        [<Chain id=A>]

        Now, we will read the ligand and append it to the existing protein
        structure.

        >>> from luna.mol.entry import MolFileEntry
        >>> mol_file = "tutorial/inputs/ZINC000007786517.mol"
        >>> e = MolFileEntry.from_mol_file("D4", "ZINC000007786517", mol_file,
        ...                                mol_obj_type='rdkit', \
is_multimol_file=False)
        >>> joined_structure = e.get_biopython_structure(structure)

        Observe that now the list of chains contains chains 'A' and 'z',
        which is the default chain where ligands are added.

        >>> print(joined_structure[0].child_list)
        [<Chain id=A>, <Chain id=z>]

        If we loop over the residues in chain 'z', we will find our ligand.

        >>> for r in structure[0]["z"]:
        >>>     print(r)
        <Residue LIG het=H_LIG resseq=9999 icode= >
        """

        if parser is None:
            parser = PDBParser(PERMISSIVE=True, QUIET=True,
                               FIX_EMPTY_CHAINS=True,
                               FIX_ATOM_NAME_CONFLICT=True,
                               FIX_OBABEL_FLAGS=False)

        mol_file_ext = self.mol_file_ext
        if mol_file_ext is None and self.mol_file is not None:
            mol_file_ext = get_file_format(self.mol_file)

        if self.mol_obj_type == "openbabel":
            pdb_block = self.mol_obj.to_pdb_block()

            atm = self.mol_obj.unwrap().GetFirstAtom()
            residue_info = atm.GetResidue()

            # When the PDBParser finds an empty chain, it automatically
            # replace it by 'z'.
            chain_id = (residue_info.GetChain()
                        if residue_info.GetChain().strip() != ""
                        else self.chain_id)
            comp_num = residue_info.GetNum()

            if residue_info.GetName() in WATER_NAMES:
                comp_name = "W"
            elif residue_info.IsHetAtom(atm):
                comp_name = "H_%s" % residue_info.GetName()
            else:
                comp_name = " "

            if mol_file_ext == "pdb":
                self.chain_id = chain_id
                self.comp_name = residue_info.GetName()
                self.comp_num = comp_num
                self.is_hetatm = residue_info.IsHetAtom(atm)
        else:
            pdb_block = self.mol_obj.to_pdb_block()

            if mol_file_ext == "pdb":
                residue_info = (self.mol_obj.unwrap()
                                .GetAtoms()[0].GetPDBResidueInfo())
                # When the PDBParser finds an empty chain, it automatically
                # replace it by 'z'.
                chain_id = (residue_info.GetChainId()
                            if residue_info.GetChainId().strip() != ""
                            else self.chain_id)
                comp_num = residue_info.GetResidueNumber()

                if residue_info.GetResidueName() in WATER_NAMES:
                    comp_name = "W"
                elif residue_info.GetIsHeteroAtom():
                    comp_name = "H_%s" % residue_info.GetResidueName()
                else:
                    comp_name = " "

                self.chain_id = chain_id
                self.comp_name = residue_info.GetResidueName()
                self.comp_num = comp_num
                self.is_hetatm = residue_info.GetIsHeteroAtom()
            else:
                # When the PDBParser finds an empty chain,
                # it automatically replace it by 'z'.
                chain_id = self.chain_id
                comp_name = "H_UNL"
                comp_num = 1

        comp_structure = parser.get_structure_from_pdb_block(self.pdb_id,
                                                             pdb_block)

        chain = comp_structure[0][chain_id]
        if self.chain_id != chain.id:
            chain.id = self.chain_id

        lig = chain[(comp_name, comp_num, " ")]

        # It only substitutes the ligand id if it is different from the id
        # defined by the MolFileEntry object property. This update will never
        # happen when the ligand file is a PDB file as the ids are guaranteed
        # to be equal.
        if lig.id != ("H_%s" % self.comp_name, self.comp_num, " "):
            lig.id = ("H_%s" % self.comp_name, self.comp_num, " ")

        lig.resname = self.comp_name

        if entity is not None:
            if isinstance(entity, Entity):
                structure = entity.get_parent_by_level('S')
                if self.chain_id not in structure[0].child_dict:
                    chain = chain.copy()
                    structure[0].add(chain)
                else:
                    lig = lig.copy()
                    # Update the ligand index according to the number of
                    # molecules already present in the chain.
                    lig.idx = len(structure[0][self.chain_id].child_list)
                    structure[0][self.chain_id].add(lig)
            else:
                raise IllegalArgumentError("The informed entity is not a "
                                           "valid Biopython object.")
        else:
            entity = comp_structure

        return entity

    def __repr__(self):
        return '<MolFileEntry: %s%s%s>' % (self.pdb_id, self.sep, self.mol_id)

    def __getstate__(self):
        if self._mol_obj is not None:
            self.mol_obj = MolWrapper(self.mol_obj)
        return self.__dict__

    def __setstate__(self, state):
        self.__dict__.update(state)


def recover_entries_from_entity(entity, get_small_molecules=True,
                                get_chains=False, ignore_artifacts=True,
                                by_cluster=False, sep=ENTRY_SEPARATOR):
    """ Search for chains and small molecules in ``entity`` and return
    them as strings.

    Parameters
    ----------
    entity : :class:`~luna.pdb.core.entity.Entity`
        An entity from where chains or small molecules will be recovered.
    get_small_molecules : bool
        If True, identify small molecules and return them as `MolEntry`
        objects. The default value is True.
    get_chains : bool
        If True, identify chains and return them as `ChainEntry` objects.
        The default value is False.
    ignore_artifacts : bool
        If True, ignore the following crystallography artifacts: ACE, ACT,
        BME, CSD, CSW, EDO, FMT, GOL, MSE, NAG, NO3, PO4, SGM, SO4, or
        TPO. The default value is True.
    by_cluster : bool
        If True, aggregate entries by cluster. Cluster ids are exclusive
        to :class:`~luna.pdb.core.residue.Residue` instances and are
        automatically set by :class:`~luna.pdb.parser.ftmap.FTMapParser`,
        a parser for FTMap results. By default, the cluster id of
        :class:`~luna.pdb.core.residue.Residue` instances are set to None,
        therefore, if the cluster id is not explicitly defined, all entries
        will be aggregated to the same key ``None``.
    sep : str
        A separator character to format the entry string.
        The default value is ':'.

    Returns
    ------
     : list or dict
        If ``by_cluster`` is set to False, a list of `ChainEntry` or
        `MolEntry` objects is returned. Otherwise, a dict is returned,
        in which keys are clusters and values are lists of `ChainEntry`
        or `MolEntry` objects. When no cluster information is available, all
        entries are aggregated in a key of value ``None``. Cluster ids are
        exclusive to :class:`~luna.pdb.core.residue.Residue` instances,
        therefore, `ChainEntry` objects are always placed in a key of
        value ``None``.

    Examples
    --------

    First, let's parse a PDB file.

    >>> from luna.pdb.parser.base.PDBParser import PDBParser
    >>> pdb_parser = PDBParser(PERMISSIVE=True, QUIET=True)
    >>> pdb_file = "tutorial/inputs/3QQK.pdb"
    >>> structure = pdb_parser.get_structure("Protein", pdb_file)

    Now, we can recover entries from the parsed PDB file:

    >>> from luna.mol.entry import recover_entries_from_entity
    >>> entries = recover_entries_from_entity(structure, get_chains=True)
    >>> for e in entries:
    >>>     print(e)
    <MolEntry: Protein:A:X02:497>
    <ChainEntry: Protein:A>

    """

    clusters = defaultdict(list)
    entries = []

    if entity.level == "S":
        if get_small_molecules:
            residues = entity[0].get_residues()
        if get_chains:
            chains = entity[0].get_chains()

    elif entity.level == "M":
        if get_small_molecules:
            residues = entity.get_residues()
        if get_chains:
            chains = entity.get_chains()
    else:
        if get_small_molecules:
            # If the entity is already a Chain, get_parent_by_level() returns
            # the same object. But, if the entity is a Residue or Atom, it will
            # return its corresponding chain parent.
            residues = entity.get_parent_by_level("C").get_residues()
        if get_chains:
            chains = entity.get_parent_by_level("M").get_chains()

    if get_small_molecules:
        pdb_id = entity.get_parent_by_level("S").id
        for res in residues:
            if res.is_hetatm():
                if ignore_artifacts and res.resname in ARTIFACTS_LIST:
                    continue

                comp_num_and_icode = ""
                if isinstance(res.id[1], int):
                    comp_num_and_icode = str(res.id[1])
                comp_num_and_icode += \
                    str(res.id[2]) if res.id[2].strip() else ""

                entry = MolEntry.from_string(sep.join([pdb_id, res.parent.id,
                                                       res.resname,
                                                       comp_num_and_icode]))
                if by_cluster:
                    clusters[res.cluster_id].append(entry)
                else:
                    entries.append(entry)

    if get_chains:
        pdb_id = entity.get_parent_by_level("S").id
        for chain in chains:
            entry = ChainEntry.from_string(sep.join([pdb_id, chain.id]))
            if by_cluster:
                clusters[None].append(entry)
            else:
                entries.append(entry)

    if by_cluster:
        return clusters
    return entries
